// system.C


#include "rkbp.h"

extern  double  hybrid_factor;
extern  ulong   mask[33];
extern  long    pack_size[12];
extern  long    print_level;


// The following macro is used in the input functions to check
// that a string str read in agrees with the string  s.

#define check(s) if ( 0 != strcmp(str,s) ) goto invalid_input


vector_long   balance_vector(0);
vector_long   heuristic_vector(0);
rule          heuristic_rule;



/////////////////////////////////////////////////////////////////
//
//             rewriting_system::rewriting_system
//
/////////////////////////////////////////////////////////////////

rewriting_system::rewriting_system(void)
: autn(50),  // '50' is arbitrary!
  system_file()
{
  a_type      = MONOID;
  patterns    = FALSE;
  o_type      = LNLX;
  use_heuristics = TRUE;
  use_broken  = FALSE;
  n_gens      = 0;
  t_gens      = 1;
  n_xgens     = 0;
  t_xgens     = 1;
  n_bits      = 0;
  gens        = NULL;
  n_sub       = 0;
  rw_strategy = FROM_LEFT;
  rewrite_verbose = FALSE;
  use_patterns    = FALSE;
//  pRewrite    = NULL;
  kb_strategy = TWO_STEP;
//  pKnuth_Bendix = kb_two_step;
  compare     = LNLX;
  savell      = LONG_MAX;
  savelr      = LONG_MAX;
  bsize       = 1000;
  logll       = 0;
  n_rewrite   = 0;
  total_input = 0;
  n_applications = 0;
  total_left     = 0;
  noverlaps      = 0;
  nbalance       = 0;
  nsave          = 0;
  nsave_total    = 0;
  nsave_max      = 0;
  n_rabin_karp   = 0;
  n_rk_tries     = 0;
  n_rk_hits      = 0;

  first_subsystem = NULL;
  last_subsystem  = NULL;
  first_save      = NULL;
  last_save       = NULL;
  broken_subword_rules = NULL;
  broken_prefix_rules  = NULL;
}



/////////////////////////////////////////////////////////////////
//
//           rewriting_system::~rewriting_system
//
/////////////////////////////////////////////////////////////////

rewriting_system::~rewriting_system(void)
{
  subsystem  *subsyst,*next;

  if (logll > 0) log_stream.close();

  delete[] gens;

  subsyst = first_subsystem;

  while (subsyst != NULL) {
    next = subsyst->get_next_subsystem();
    delete subsyst;
    subsyst = next;
  }

  if (broken_prefix_rules != NULL)
    delete broken_prefix_rules;

  if (broken_subword_rules != NULL)
    delete broken_subword_rules;
}



/////////////////////////////////////////////////////////////////
//
//               rewriting_system::new_subsystem
//
/////////////////////////////////////////////////////////////////

subsystem *rewriting_system::new_subsystem(
  char  *str
)
{
  subsystem   *subsyst;
  
  subsyst = new subsystem;

  if (str != NULL) subsyst->set_subsystem_file(str);
  subsyst->set_parent(this);

  return subsyst;
}



/////////////////////////////////////////////////////////////////
//
//                  rewriting_system::input
//
/////////////////////////////////////////////////////////////////

// The rewriting system is assumed to be freshly initialized by
// the default constructor.

void rewriting_system::input(
  char   *file_name,
  bool     check
)
{
  long      exists,lname,i;
  char      file[100];
  char      directory[100];
  bool      echo;
  ifstream  stream;

  echo = get_echo();

  lname = strlen(file_name);
  
  if (lname > 90) {
    cout << "Name too long in rewriting_system::input" << endl;
    return;
  }
    
  strcpy(file,file_name);
  strcat(file,".sys");
  exists = (lname > 0) &&
           (0 == strcmp(file_name,"none") || file_exists(file) );
  
  if (!exists && lname > 0) {
    cout << "file " << file << " does not exist." << endl;
    if (echo) exit(1);
  }
  
  while (!exists) {
    cout <<
    "\nEnter problem name (or \"none\" or \"stop\"):  ";
    scan_string(cin,file);

    if (0 == strcmp(file,"stop") ) return;
    if (0 == strcmp(file,"none") ) break;
    strcat(file,".sys");

    exists = file_exists(file);

    if (!exists) {
      cout << "file does not exist." << endl;
      if (echo) exit(1);
    }
  }

  if ( 0 == strcmp(file,"none") ) {
    system_file.set_data("");
    directory[0] = '\0';
    input(directory,cin,TRUE,check);
  }

  else {
    system_file.set_data(file);
    strcpy(directory,file);

    i = strlen(directory)+1;
    while (i > 0 && directory[i-1] != '/') i--;
    directory[i] = '\0';

    stream.open(file, ios::in);
    input(directory,stream,FALSE,check);
  }
}



void rewriting_system::input(
  char*     directory,
  istream  &stream,
  bool      prompt,
  bool      check
)
{
  long                  i,j,k,ninv,nsub,m;
  char                  first_gen;
  char                  str[500],str2[500],pfx[10];
  bool                  echo;
  rule_type             rtype;
  vector_char           vec;
  vector_long           wrd;
  subsystem            *subsyst;
  naming_convention     name_conv;
  numbering_convention  numb_conv;
  inverse_convention    inv_conv;
  convention            wt_conv;
  convention            lev_conv;

  echo = get_echo();

 get_patterns:
  if (prompt) {
    cout << "\nWill rule patterns be used (y/n):  ";

    if (YES) patterns = TRUE;
    else     patterns = FALSE;
  }

 get_word_type:
  if (prompt) {
    cout << "\nEnter word_type (monoid or group):  ";
    scan_string(stream,str);
  }

  else {
    stream >> str;

    if      (0 == strcmp(str,"system") )  patterns = FALSE;
    else if (0 == strcmp(str,"xsystem") ) patterns = TRUE;
    else goto invalid_input;

    stream >> str;
    check("word_type:");
    stream >> str;
  }

  if      ( 0 == strcmp(str,"monoid") ) a_type = MONOID;
  else if ( 0 == strcmp(str,"group") )  a_type = GROUP;

  else {
    cout << "Invalid word_type" << endl;
    if (prompt && !echo) goto get_word_type;
    exit(1);
  }

  if (prompt) {
    cout << "\nEnter ngens:  ";
    scan_long(stream,n_gens);

    if (patterns) {
      cout << "\nEnter nxgens:  ";
      scan_long(stream,n_xgens);
    }

    else n_xgens = 0;
  }

  else {
    stream >> str;
    check("ngens:");
    stream >> n_gens;

    if (patterns) {
      stream >> str;
      check("nxgens:");
      stream >> n_xgens;
    }

    else n_xgens = 0;
  }

  n_xgens += n_gens;

  if (a_type == MONOID) {
    t_gens  = n_gens;
    t_xgens = n_xgens;
  }

  else {
    t_gens  = 2*n_gens;
    t_xgens = 2*n_xgens;
  }
  
  m = 2;
  n_bits = 1;
  
  while (m < t_xgens) {
    n_bits++;
    m *= 2;
  }

  gens = new generator[max(t_xgens,2*t_gens)];

  // Get naming convention.

 get_name_conv:
  if (prompt) {
    cout <<
    "\nEnter naming_convention, one of the following:\n" <<
    "prefix <string>, letters <letter>, or none:  ";
    stream >> str;
    if (echo) cout << str;
  }

  else {
    stream >> str;
    check("naming_convention:");
    stream >> str;
  }


  if (0 == strcmp(str,"prefix") ) {
    name_conv = PREFIX;

    if (prompt) {
      if (echo) cout << " ";
      scan_string(stream,pfx);
    }

    else stream >> pfx;
  }

  else if (0 == strcmp(str,"letters") ) {
    name_conv = LETTERS;

    if (prompt) {
      if (echo) cout << " ";
      scan_string(stream,str2);
    }

    else stream >> str2;

    if (strlen(str2) != 1) {
      cout << "letter convention requires a single letter"
           << endl;

      if (prompt && !echo) goto get_name_conv;
      else exit(1);
    }

    first_gen = *str2;

    if ( !islower(first_gen) || !islower(first_gen+n_xgens-1) ) {
      cout << "invalid letter generator" << endl;

      if (prompt && !echo) goto get_name_conv;
      else exit(1);
    }
  }

  else {
    if (prompt && echo) cout << endl;
    if (0 == strcmp(str,"none") ) name_conv = NO_NAME;

    else {
      cout << "Invalid naming convention" << endl;
      if (prompt && !echo) goto get_name_conv;
      exit(1);
    }
  }

  if (a_type == GROUP) {
    // Get numbering convention.

   get_numb_conv:
    if (prompt) {
      cout <<
        "\nEnter numbering_convention (1_2, 1_-1, or none):  ";
      scan_string(stream,str);
    }

    else {
      stream >> str;
      check("numbering_convention:");
      stream >> str;
    }

    if      (0 == strcmp(str,"1_-1") ) numb_conv = a_A;


    else if (0 == strcmp(str,"1_2") ) {
      if (patterns) {
        cout << "Patterns don't permit 1_2 convention!" << endl;
        if (prompt && !echo) goto get_numb_conv;
        exit(1);
      }

      numb_conv = a_b;
    }

    else if (0 == strcmp(str,"none") ) {
      if (name_conv != NO_NAME) {
        cout << "Numbering_convention requires naming convention"
             << endl;
        if (prompt && !echo) return;
        exit(1);
      }

      numb_conv = NO_NUMB;
    }

    else {
      cout << "Invalid numbering convention" << endl;
      if (prompt && !echo) goto get_numb_conv;
      exit(1);
    }
      

    // Get inverse convention.

   get_inv_conv:
    if (prompt) {
      cout << endl <<
        "Enter inverse_convention (case, exponent, or none):  ";
      scan_string(stream,str);
    }

    else {
      stream >> str;
      check("inverse_convention:");
      stream >> str;
    }

    if      (0 == strcmp(str,"case") )     inv_conv = CASE;
    else if (0 == strcmp(str,"exponent") ) inv_conv = EXPONENT;

    else if (0 == strcmp(str,"none") ) {
      if (name_conv != NO_NAME) {
        cout << "Inverse_convention requires naming convention"
             << endl;

        if (prompt && !echo) return;
        exit(1);
      }

      inv_conv = NO_INV;
    }

    else {
      cout << "Invalid inverse convention" << endl;
      if (prompt && !echo) goto get_inv_conv;
      exit(1);
    }

    // If a naming convention was specified, we have enough
    // information to create the generator list.

    if (name_conv != NO_NAME) {
      for (i = 0; i < n_xgens; i++) {
        if (numb_conv == a_b) {
          j = i;
          k = i + n_xgens;
        }

        else {
          j = 2*i;
          k = j + 1;
        }

        if (name_conv == PREFIX) sprintf(str,"%s%ld",pfx,i+1);

        else {
          str[0] = first_gen + i;
          str[1] = '\0';
        }

        strcpy(str2,str);

        if (inv_conv == CASE) other_case(str2);
        else strcat(str2,"^-1");

        gens[j].set_name(str);
        gens[j].set_inv_numb(k);
        gens[k].set_name(str2);
        gens[k].set_inv_numb(j);
      }
    }  // if (name_conv != NO_NAME)

    else {   // We need more information.
      if (numb_conv == NO_NUMB) {
        if (prompt) {
          cout <<
          "\nEnter the generator names, including inverses,\n" <<
          "in increasing order:\n";
        }

        else {
          stream >> str;
          check("generators:");
        }

        for (i = 0; i < t_xgens; i++) {
          stream >> str;
          gens[i].set_name(str);
          if (prompt && echo) cout << str << endl;
        }

        if (prompt) stream.ignore(10000,'\n');
      }

      if (inv_conv == NO_INV ) {
       get_inv_pairs:
        if (prompt)
          cout <<
          "\nEnter inverse_pairs of generators:\n";

        else {
          stream >> str;
          check("inverse_pairs:");
        }

        for (i = 0; i < n_xgens; i++) {
          stream >> str >> str2;

          if (prompt && echo) cout << str << " " << str2 << endl;

          if (numb_conv == NO_NUMB) {
            j = gen_number(str);
            k = gen_number(str2);

            if (j < 0 || k < 0) {
              cout << "Invalid inverse pair" << endl;
              if (prompt && !echo) goto get_inv_pairs;
              exit(1);
            }

            gens[j].set_inv_numb(k);
            gens[k].set_inv_numb(j);
          }

          else if (numb_conv == a_b) {
            j = i + n_xgens;
            gens[i].set_name(str);
            gens[i].set_inv_numb(j);
            gens[j].set_name(str2);
            gens[j].set_inv_numb(i);
          }

          else if (numb_conv == a_A) {
            j = 2*i;
            gens[j].set_name(str);
            gens[j].set_inv_numb(j+1);
            gens[j+1].set_name(str2);
            gens[j+1].set_inv_numb(j);
          }
        }

        if (prompt) stream.ignore(10000,'\n');
      }

      else if (numb_conv != NO_NUMB) {
        if (prompt) {
          cout <<
          "\nEnter the generator names, without inverses,\n" <<
          "in increasing order:\n";
        }

        else {
          stream >> str;
          check("generators:");
        }

        for (i = 0; i < n_xgens; i++) {
          stream >> str;
          if (prompt && echo) cout << str << endl;
          strcpy(str2,str);

          if (inv_conv == CASE) other_case(str2);
          else (void) strcat(str2,"^-1");

          if (numb_conv == a_b) {
            j = i;
            k = i + n_xgens;
          }

          else {
            j = 2*i;
            k = j + 1;
          }

          gens[j].set_name(str);
          gens[j].set_inv_numb(k);
          gens[k].set_name(str2);
          gens[k].set_inv_numb(j);
        }

        if (prompt) stream.ignore(10000,'\n');

      }  // if (numb_conv != NO_NUMB)

      else {  // numb_conv == NO_NUMB and inv_conv != NO_INV
        for (i = 0; i < t_xgens; i++) {
          if (gens[i].get_inv_numb() >= 0) continue;

          gens[i].get_name(str);
          strcpy(str2,str);

          if (inv_conv == CASE) other_case(str2);
          else strcat(str2,"^-1");

          if (i >= (j = gen_number(str2) ) ) {
            cout << "Unable to find inverse with name "
                 << str2 << endl;

            if (prompt && !echo) return;
            exit(1);
          }

          gens[i].set_inv_numb(j);
          gens[j].set_inv_numb(i);
        }
      }  // numb_conv == NO_NUMB and inv_conv != NO_INV

    }  // No naming convention

    // Construct the inverse words.

    for (i = 0; i < t_xgens; i++) {
      j = gens[i].get_inv_numb();
      gens[i].set_invertible(TRUE);
      gens[i].set_cancelable(TRUE);
      gens[i].set_one(j);
    }

    make_generator_index();

  }  // if (a_type == GROUP)

  else {  // MONOID words
    if (name_conv != NO_NAME) {
      for (i = 0; i < n_xgens; i++) {
        if (name_conv == PREFIX) sprintf(str,"%s%ld",pfx,i+1);

        else {
          str[0] = first_gen + i;
          str[1] = '\0';
        }

        gens[i].set_name(str);
      }
    }  // if (name_conv != NO_NAME)

    else {  // name_conv == NO_NAME
      if (prompt)
        cout << "\nEnter generator names in increasing order:\n";

      else {
        stream >> str;
        check("generators:");
      }

      for (i = 0; i < n_xgens; i++) {
        stream >> str;
        if (prompt && echo) cout << str << endl;
        gens[i].set_name(str);
      }
    }

    make_generator_index();

    // Get inverse data.

   get_ninv:
    if (prompt) {
      cout <<
      "\nEnter number of generators with inverses:  ";
      scan_long(stream,ninv);
    }

    else {
      stream >> str;
      check("ninv:");
      stream >> ninv;
    }

    if (ninv < 0) {
      cout << "ninv must be nonnegative" << endl;
      if (prompt && !echo) goto get_ninv;
      exit(1);
    }

    if (ninv > 0) {
      if (prompt) cout << "\nEnter inverse pairs:\n";

      else {
        stream >> str;
        check("inverse_pairs:");
      }

      for (i = 0; i < ninv; i++) {
        for (;;) {
          stream >> str >> vec;  // This assumes inverses
                                 // are short!
          if (prompt && echo) cout << str << " " << vec << endl;

          if (0 > (j = gen_number(str) ) ) 
            cout << "Can't find generator with name " << str
                 << endl;

          else if (parse(vec,wrd) )
            cout << "Can't parse " << vec << endl;
          else break;

          if (!prompt || echo) exit(1);
          cout << "Reenter inverse pair:  ";
        }

        gens[j].set_invertible(TRUE);
        gens[j].set_cancelable(TRUE);
        gens[j].set_inverse(wrd);
      }
    }
  }  // else {  // MONOID word

  // Get weight data.

 get_weight_data:
  if (prompt) {
    cout << "\nEnter weight_convention (constant or none):  ";
    scan_string(stream,str);
  }

  else {
    stream >> str;
    check("weight_convention:");
    stream >> str;
  }

  if (0 == strcmp(str,"constant") ) {
    wt_conv = CONSTANT;
    for (i = 0; i < t_xgens; i++) gens[i].set_weight(1);
  }

  else if (0 == strcmp(str,"none") ) {
    wt_conv = NONCONSTANT;
    
    if (prompt)
      cout << "\nEnter the weights of generators,\n" <<
      "in increasing order of the generators:\n";

    else {
      stream >>str;
      check("weights:");
    }

    for (i = 0; i < t_xgens; i++) {
      stream >> j;

      if (prompt && echo) cout << j << endl;

      for (;;) {
        if (j > 0) break;
        cout << "Weights must be positive" << endl;
        if (!prompt || echo) exit(1);
        cout << "Reenter last weight:  ";
        stream >> j;
      }

      gens[i].set_weight(j);
    }
  }

  else {
    cout << "Invalid weight convention" << endl;
    if (prompt && !echo) goto get_weight_data;
    exit(1);
  }

  // Get level data.

 get_level_data:
  if (prompt) {
    cout << "\nEnter the level_convention\n" <<
      "(constant, (r)basic_wreath,\n" <<
      "(r)inverse_pair_wreath, or (r)none):  ";

    scan_string(stream,str);
  }

  else {
    stream >> str;
    check("level_convention:");
    stream >> str;
  }

  if (0 == strcmp(str,"constant") ) {
    lev_conv = CONSTANT;
    for (i = 0; i < t_xgens; i++) gens[i].set_level(1);
  }

  else if (0 == strcmp(str,"basic_wreath") ||
           0 == strcmp(str,"rbasic_wreath")  ) {
    if (0 == strcmp(str,"basic_wreath") )lev_conv = NONCONSTANT;
    else lev_conv = RNONCONSTANT;

    if (a_type == MONOID)
      for (i = 0; i < t_xgens; i++) gens[i].set_level(i+1);

    else {
      k = 1;

      for (i = 0; i < t_xgens; i++) {
        j = gens[i].get_inv_numb();
        if (j <= i) continue;
        gens[j].set_level(k++);
        gens[i].set_level(k++);
      }
    }
  }

  else if (0 == strcmp(str,"inverse_pair_wreath") ||
           0 == strcmp(str,"rinverse_pair_wreath")  ) {
    if (0 == strcmp(str,"inverse_pair_wreath") )
      lev_conv = NONCONSTANT;
    else
      lev_conv = RNONCONSTANT;
    
    if( a_type != GROUP || numb_conv != a_A) {
      cout << "inverse_pair_wreath convention not allowed here."
           << endl;
      if (prompt && !echo) goto get_level_data;
      exit(1);
    }

    for (i = j = 0; i < n_xgens; i++) {
      gens[j++].set_level(i+1);
      gens[j++].set_level(i+1);
    }
  }

  else if (0 == strcmp(str,"none") ||
           0 == strcmp(str,"rnone")  ) {
    if (0 == strcmp(str,"none") ) lev_conv = NONCONSTANT;
    else lev_conv = RNONCONSTANT;
    
    if (prompt)
      cout << "\nEnter the levels of generators,\n" <<
        "in increasing order of the generators:\n";

    else {
      stream >> str;
      check("levels:");
    }

    for (i = 0; i < t_xgens; i++) {
      stream >> j;

      if (prompt && echo) cout << j << endl;

      for (;;) {
        if (j > 0) break;
        cout << "Levels must be positive." << endl;
        if (!prompt || echo) exit(1);
        cout << "Reenter last level:  ";
        stream >> j;
      }

      gens[i].set_level(j);
    }
  }

  else {
    cout << "Unknown level convention" << endl;
    if (prompt && !echo) goto get_level_data;
    exit(1);
  }
  
  use_heuristics = FALSE;

  if ( wt_conv == CONSTANT)
    if (lev_conv == CONSTANT) {
      set_o_type(LNLX);
      use_heuristics = TRUE;
    }

    else if (lev_conv == NONCONSTANT) set_o_type(WRLNLX);
    else set_o_type(RWRLNLX);

  else
    if (lev_conv == CONSTANT) set_o_type(WTLNLX);
    else if (lev_conv == NONCONSTANT) set_o_type(WRWTLNLX);
    else set_o_type(RWRWTLNLX);

  // Get the subsystems.

  if (prompt) {
    cout << "\nEnter the number of subsystems:  ";
    scan_long(stream,nsub);
  }

  else {
    stream >> str;
    check("nsub:");
    stream >> nsub;
  }

  if (nsub <= 0) {
    if ( !stream.ignore(10000,'\n') ) {
      cout << "Error clearing command line." << endl;
      exit(1);
    }

    cout << endl;
    return;
  }

  if (!prompt) {
    stream >> str;
    check("subsystems:");
  }

  for (i = 0; i < nsub; i++) {
    subsyst = new subsystem;
    subsyst->set_parent(this);
    subsyst->input(directory,stream,prompt,check);

    if (NO_INDEX == subsyst->get_i_type() ) {
      rtype = subsyst->get_r_type();

      if (PREFIX_RULE == rtype)
        if (broken_prefix_rules == NULL)
          broken_prefix_rules = subsyst;
        else goto broken;

      else if (SUBWORD_RULE == rtype)
        if (broken_subword_rules == NULL)
          broken_subword_rules = subsyst;
        else goto broken;

      else if (PREFIX_PATTERN  == rtype ||
               SUBWORD_PATTERN == rtype   )
        add_subsystem(subsyst);
    }

    else add_subsystem(subsyst);
  }

  if (patterns) {
    cout <<
      "Note that patterns will not be used during rewriting\n"
         <<
      "unless the command ""patterns"" is issued." << endl;
  }

  return;

 invalid_input:
  cout << "Invalid input string \"" << str << "\"" << endl;
  exit(1);

 broken:
  cout << "Too many subsystems with no indexes." << endl;

  return;
}



/////////////////////////////////////////////////////////////////
//
//              rewriting_system::add_subsystem
//
/////////////////////////////////////////////////////////////////

void rewriting_system::add_subsystem(
  subsystem  *subsyst
)
{
  subsyst->set_next_subsystem(NULL);
  subsyst->set_prev_subsystem(last_subsystem);

  if ( last_subsystem != NULL) {
    last_subsystem->set_next_subsystem(subsyst);
    last_subsystem = subsyst;
  }

  else
    first_subsystem = last_subsystem = subsyst;

  n_sub++;

  return;
}



/////////////////////////////////////////////////////////////////
//
//            rewriting_system::make_generator_index
//
/////////////////////////////////////////////////////////////////

void rewriting_system::make_generator_index()
{
  long   i;
  char   str[500];

  for (i = 0; i < t_xgens; i++) {
    gens[i].get_name(str);
    autn.enter_string(str,i);
  }

  return;
}



/////////////////////////////////////////////////////////////////
//
//                 rewriting_system::gen_number
//
/////////////////////////////////////////////////////////////////

long rewriting_system::gen_number(
  char  *str
)
{
  long  i;

  for (i = 0; i < t_xgens; i++)
    if ( gens[i].compare_name(str) ) return i;

  return -1;
}



/////////////////////////////////////////////////////////////////
//
//               rewriting_system::parse
//
/////////////////////////////////////////////////////////////////

long rewriting_system::parse(
  vector_char  &vec,
  vector_long  &wrd
)
{
  long    lvec,j,g,state,current,p,bit,h,old;
  char    c;
  
  lvec = vec.get_length();
  wrd.set_length(0);

  if (lvec == 0 || lvec == 1 && vec[0] == '@') return 0;

  j       = 0;
  g       = LONG_MAX;
  state   = 0;
  current = 0;  

  while (current < lvec) {
    c = vec[current++];

    for (p = CHAR_BIT-2; p >= 0; p--) {
      bit   = (c>>p) & 1;
      state = autn.get_entry(state,bit);
      if (state == LONG_MAX) break;
    }

    if ( LONG_MAX != state &&
         LONG_MAX != (h = autn.get_entry(state,2) ) ) {
      //  The current prefix is a generator name.

      g   = h;
      old = current;
    }

    if (state == LONG_MAX || current >= lvec)
      if (g != LONG_MAX) {
        if (g >= t_xgens) {
          cout << "Invalid generator in parse" << endl;
          return 1;
        }

        current = old;
        state   = 0;
        wrd.safe_set_entry(j++,g);
      }

      else {
        cout << "Unable to parse names string" << endl;
        return 1;
      }
  }

  return 0;
}



/////////////////////////////////////////////////////////////////
//
//                  rewriting_system::parse_6bit
//
/////////////////////////////////////////////////////////////////

long rewriting_system::parse_6bit(
  vector_char  &vec,
  long          nbits,
  long          len,
  vector_long  &word
)
{
  long   i,wbits,zbits,j,diff,bits_in_vec,bits_to_save;
  ulong  z,w;

  word.set_length(len);
  bits_in_vec = len*nbits;
  bits_to_save = len*nbits;
  i = j = 0;
  w = z = 0;
  wbits = zbits = 0;

  while (bits_to_save > 0) {
    while (zbits < nbits) {
      diff = nbits - zbits;

      if (wbits == 0 && bits_in_vec > 0) {
        w = vec[i++] - 48;
        w = w&mask[6];
        wbits = 6;
        bits_in_vec -= 6;
      }

      if (wbits > diff) {
        z = (z<<diff) | ((w>>(wbits-diff))&mask[diff]);
        zbits  = nbits;
        wbits -= diff;
      }

      else if (wbits > 0) {
        z = (z<<wbits) | (w&mask[wbits]);
        zbits += wbits;
        wbits  = 0;
      }

      else {
        z = z<<(nbits-zbits);
        zbits = nbits;
      }
    }

    if (z >= t_xgens) {
      cout << "Invalid generator in parse_6bit" << endl;
      cout << "Generator " << z << " for position " << i << endl;
      cout << vec << endl;
      return 1;
    }

    word.set_entry(j++,z);
    bits_to_save -= nbits;
    zbits = z = 0;
  }

  return 0;
}



/////////////////////////////////////////////////////////////////
//
//                      set_hybrid_factor
//
/////////////////////////////////////////////////////////////////

void set_hybrid_factor(
  double  factor
)
{
  hybrid_factor = factor;
}



/////////////////////////////////////////////////////////////////
//
//                  rewriting_system::read_word
//
/////////////////////////////////////////////////////////////////

// Returns 0 if successful and 1 if end of file has been reached
// already.  If an error or end of file condition occurs in the
// middle of reading the word, the current input line is cleared
// and a value of 2 returned.

long rewriting_system::read_word(
  istream        &stream,
  external_form   exf,
  long            nbits,
  vector_long    &wrd
)
{
  long                lwx,i,g,mxl,code,start;
  static vector_char  buff,buff1;

  if (exf == LEN_NAMES   ||
      exf == LEN_NUMBERS ||
      exf == LEN_6BIT      ) {
    if ( !(stream >> lwx) ) return 1;
    mxl = wrd.get_max_length();
    wrd.set_max_length(max(lwx,mxl) );
  }

  else lwx = -1;

  if (exf == LEN_NUMBERS) {
    for (i = 0; i < lwx; i++) {
      if ( !(stream >> g) ) goto input_error;
      wrd.safe_set_entry(i,g);
    }

    wrd.set_length(lwx);
    return 0;
  }

  if (exf == LEN_6BIT) {
    if ( !(stream >> buff) ) goto input_error;

    if (parse_6bit(buff,nbits,lwx,wrd) ) goto input_error;
    return 0;
  }

  if (exf == NAMES || exf == LEN_NAMES) {
    code = continued_string(stream,buff,buff1);

    if (1 == code) return 1;
    if (2 == code) goto input_error;

    start = 0;
    code = decode(buff,start,wrd);
    if ('\0' != buff[start]) code = 2;
    return code;
  }

 input_error:
  cout << "Input error in read_word." << endl;

  if ( !stream.ignore(10000,'\n') ) {
    cout << "Unable to clear current line" << endl;
    exit(1);
  }

  return 2;
}



/////////////////////////////////////////////////////////////////
//
//                 rewriting_system::summary
//
/////////////////////////////////////////////////////////////////

void rewriting_system::summary(void)
{
  subsystem   *subsyst;

  cout << "\nalphabet_type:  ";

  if (a_type == MONOID) cout << "monoid";
  else cout << "group";

  cout << ",  ngens:  " << n_gens << ",   nbits:  " << n_bits
       << endl;

  if (patterns) cout << "n_xgens:  " << (n_xgens-n_gens) << endl;

  cout << "rewriting strategy:  ";

  if (rw_strategy == FROM_LEFT) cout << "from_left" << endl;
  else cout << "global" << endl;
  
  cout << "Knuth-Bendix strategy:  ";
  
  if (kb_strategy == BASIC) cout << "basic" << endl;
  else cout << "two_step" << endl;
  
  if (n_sub == 1) cout << "There is 1 subsystem:";
  else
    cout << "There are " << n_sub << " subsystems:";

  cout << endl;

  subsyst = first_subsystem;

  while (subsyst != NULL) {
    subsyst->summary();
    subsyst = subsyst->get_next_subsystem();
  }

  if (broken_prefix_rules != NULL) {
    cout << "\nThere is a subsystem of broken prefix rules."
         << endl << endl;
    broken_prefix_rules->summary();
  }

  if (broken_subword_rules != NULL) {
    cout << "\nThere is a subsystem of broken subword rules."
         << endl << endl;
    broken_subword_rules->summary();
  }
}



/////////////////////////////////////////////////////////////////
//
//                 rewriting_system::word_weight
//
/////////////////////////////////////////////////////////////////

long rewriting_system::word_weight(
  vector_long   &wrd
)
{
  long   wt,len,i;

  wt  = 0;
  len = wrd.get_length();
  for (i = 0; i < len; i++) wt += get_weight(wrd[i]);

  return wt;
}



/////////////////////////////////////////////////////////////////
//
//                   rewriting_system::lnlx
//
/////////////////////////////////////////////////////////////////

long rewriting_system::lnlx(
  vector_long  &u,
  vector_long  &v
)
{
  long   lu,lv,i,ui,vi;

  if ( (lu = u.get_length() ) != (lv = v.get_length() ) )
    if (lu < lv) return -1;
    else return 1;

  for (i = 0; i < lu; i++)
    if ( (ui = u[i]) != (vi = v[i]) )
      if (ui < vi) return -1;
      else return 1;

  return 0;
}



/////////////////////////////////////////////////////////////////
//
//                   rewriting_system::wtlnlx
//
/////////////////////////////////////////////////////////////////

long rewriting_system::wtlnlx(
  vector_long  &u,
  vector_long  &v
)
{
  long   wtu,wtv;

  if ( (wtu = word_weight(u) ) != (wtv = word_weight(v) ) )
    if (wtu < wtv) return -1;
    else return 1;

  return lnlx(u,v);
}



/////////////////////////////////////////////////////////////////
//
//                   rewriting_system::wreath
//
/////////////////////////////////////////////////////////////////

long rewriting_system::wreath(
  vector_long  &u,
  vector_long  &v
)
{
  long    lbu,lbv,su,sv,levu,levv,sgn;
  static vector_long   ubuff(100), vbuff(100);
  vector_long_segment  useg,vseg;

  level_decomposition(u,v,ubuff,vbuff);
  lbu = ubuff.get_length();
  lbv = vbuff.get_length();

  while (lbu > 0 && lbv > 0) {
    su = lbu - 1;
    sv = lbv - 1;
    levu = get_level(ubuff[su]);
    levv = get_level(vbuff[sv]);

    if (levu != levv)
      if (levu > levv) return -1;
      else return 1;

    else {
      while (su > 0 && get_level(ubuff[su-1]) == levu) su--;
      while (sv > 0 && get_level(vbuff[sv-1]) == levv) sv--;

      useg.segment(ubuff,su,lbu-su);
      vseg.segment(vbuff,sv,lbv-sv);

      if (0 != (sgn = lnlx(useg,vseg) ) )
        return sgn;

      lbu = su;
      lbv = sv;
    }
  }

  if (lbu > 0) return 1;
  if (lbv > 0) return -1;
  return 0;
}



/////////////////////////////////////////////////////////////////
//
//                   rewriting_system::rwreath
//
/////////////////////////////////////////////////////////////////

long rewriting_system::rwreath(
  vector_long  &u,
  vector_long  &v
)
{
  long    lbu,lbv,su,sv,levu,levv,sgn,tu,tv;
  static vector_long   ubuff(100), vbuff(100);
  vector_long_segment  useg,vseg;

  rlevel_decomposition(u,v,ubuff,vbuff);
  lbu = ubuff.get_length();
  lbv = vbuff.get_length();
  tu = tv = 0;

  while (lbu > tu && lbv > tv) {
    su = tu;
    sv = tv;
    levu = get_level(ubuff[su]);
    levv = get_level(vbuff[sv]);

    if (levu != levv)
      if (levu > levv) return -1;
      else return 1;

    else {
      while (su < lbu-1 && get_level(ubuff[su+1]) == levu) su++;
      while (sv < lbv-1 && get_level(vbuff[sv+1]) == levv) sv++;

      useg.segment(ubuff,tu,su-tu+1);
      vseg.segment(vbuff,tv,sv-tv+1);

      if (0 != (sgn = lnlx(useg,vseg) ) )
        return sgn;

      tu = su + 1;
      tv = sv + 1;
    }
  }

  if (lbu > tu) return 1;
  if (lbv > tv) return -1;
  return 0;
}



/////////////////////////////////////////////////////////////////
//
//             rewriting_system::level_decomposition
//
/////////////////////////////////////////////////////////////////

void rewriting_system::level_decomposition(
  vector_long  &u,
  vector_long  &v,
  vector_long  &ubuff,
  vector_long  &vbuff
)
{
  long   lu,lv,lm,lc;

  lu = u.get_length();
  lv = v.get_length();
  ubuff.set_max_length(lu);
  vbuff.set_max_length(lv);

  lm = min(lu,lv);
  lc = 0;
  while (lc < lm && u[lc] == v[lc]) lc++;

  while (lu > lc && lv > lc && u[lu-1] == v[lv-1]) {
    lu--;
    lv--;
  }

  level_sequence(u,lc,lu,ubuff);
  level_sequence(v,lc,lv,vbuff);
}



/////////////////////////////////////////////////////////////////
//
//             rewriting_system::rlevel_decomposition
//
/////////////////////////////////////////////////////////////////

void rewriting_system::rlevel_decomposition(
  vector_long  &u,
  vector_long  &v,
  vector_long  &ubuff,
  vector_long  &vbuff
)
{
  long   lu,lv,lm,lc;

  lu = u.get_length();
  lv = v.get_length();
  ubuff.set_max_length(lu);
  vbuff.set_max_length(lv);

  lm = min(lu,lv);
  lc = 0;
  while (lc < lm && u[lc] == v[lc]) lc++;

  while (lu > lc && lv > lc && u[lu-1] == v[lv-1]) {
    lu--;
    lv--;
  }

  rlevel_sequence(u,lc,lu,ubuff);
  rlevel_sequence(v,lc,lv,vbuff);
}



/////////////////////////////////////////////////////////////////
//
//              rewriting_system::level_sequence
//
/////////////////////////////////////////////////////////////////

void rewriting_system::level_sequence(
  vector_long   &u,
  long          lc,
  long          lu,
  vector_long  &ubuff
)
{
  long  lev,lbu,i,g,levg;

  lev = LONG_MAX;
  lbu = 0;

  for (i = lc; i < lu; i++) {
    g    = u[i];
    levg = get_level(g);

    if (lev >= levg) {
      ubuff.set_entry(lbu++,g);
      lev = levg;
    }
  }

  ubuff.set_length(lbu);
}



/////////////////////////////////////////////////////////////////
//
//              rewriting_system::rlevel_sequence
//
/////////////////////////////////////////////////////////////////

void rewriting_system::rlevel_sequence(
  vector_long   &u,
  long          lc,
  long          lu,
  vector_long  &ubuff
)
{
  long  lev,lbu,i,g,levg;

  lev = LONG_MAX;
  lbu = 0;

  for (i = lu-1; i >= lc; i--) {
    g    = u[i];
    levg = get_level(g);

    if (lev >= levg) {
      ubuff.set_entry(lbu++,g);
      lev = levg;
    }
  }

  ubuff.set_length(lbu);
  ubuff.reverse();
}



/////////////////////////////////////////////////////////////////
//
//                rewriting_system::set_o_type
//
/////////////////////////////////////////////////////////////////

void rewriting_system::set_o_type(
  order_type  otp
)
{
  o_type = otp;

  compare = otp;
/*
  if      (otp == LNLX)      compare = lnlx;
  else if (otp == WTLNLX)    compare = wtlnlx;
  else if (otp == WRLNLX)    compare = wrlnlx;
  else if (otp == WRWTLNLX)  compare = wrwtlnlx;
  else if (otp == RWRLNLX)   compare = rwrlnlx;
  else if (otp == RWRWTLNLX) compare = rwrwtlnlx;
*/
}



/////////////////////////////////////////////////////////////////
//
//                 rewriting_system::rewrite
//
/////////////////////////////////////////////////////////////////

void rewriting_system::rewrite(
  rule            &rul,
  rule_reference   exclude,
  bool            &changed,
  bool            &reduced
)
{
  bool       lch,rch,lrd,rrd,subword_only;
  rule_type  rtype;

  rtype = rul.r_type;

  if (PREFIX_PATTERN  == rtype ||
      SUBWORD_PATTERN == rtype   )
    error("The rewriting of patterns is not supported.");

  if (rtype == SUBWORD_RULE) subword_only = TRUE;
  else subword_only = FALSE;

  rewrite_from_left(
        subword_only,rul.left,rul.left,exclude,lch,lrd);

  rewrite_from_left(
        subword_only,rul.right,rul.right,exclude,rch,rrd);

  changed = lch||rch;
  reduced = lrd&&rrd;
}



/////////////////////////////////////////////////////////////////
//
//             rewriting_system::rewrite_from_left
//
/////////////////////////////////////////////////////////////////

void rewriting_system::rewrite_from_left(
  bool            subword_only,
  vector_long    &input,
  vector_long    &output,
  rule_reference  exclude,
  bool           &changed,
  bool           &reduced
)
{
  long         lin,lout,lout_max,bbot,lb,ll,lr,i;
  rule_type    r_type;
  packed_rule  rul;
  subsystem   *subsyst;

  // The max_length of output will not be changed.

  changed = FALSE;
  reduced = FALSE;

  lin = input.get_length();
  lout_max = output.get_max_length();

  if (lin > lout_max)
    error("output buffer too small in rewrite_from_left.");

  n_rewrite++;
  total_input += lin;

  if (rewrite_verbose) {
    cout << "Rewriting from left:  ";
    input.print();
  }

  if (n_sub <= 0) {
    output.set_copy(input);
    reduced = TRUE;
    return;
  }

  subsyst = first_subsystem;

  while (subsyst != NULL) {
    if (patterns && !use_patterns) {
      r_type = subsyst->get_r_type();

      if (r_type == SUBWORD_PATTERN || r_type == PREFIX_PATTERN) {
        subsyst = subsyst->get_next_subsystem();
        continue;
      }
    }

    subsyst->check_aux(REWRITE,subword_only,lout_max);
    subsyst = subsyst->get_next_subsystem();
  }

  lout = lin;
  bbot = lout_max;
  lb   = lin;

  while (lb > 0) output.set_entry(--bbot, input[--lb]);
  output.set_length(0);

  while (bbot < lout_max) {
    output.set_entry(lb++, output[bbot++]);
    output.set_length(lb);
    subsyst = first_subsystem;

    while (subsyst != NULL) {
      if (patterns && !use_patterns) {
        r_type = subsyst->get_r_type();

        if (r_type == SUBWORD_PATTERN || r_type == PREFIX_PATTERN) {
          subsyst = subsyst->get_next_subsystem();
          continue;
        }
      }

      subsyst->incremental_test(REWRITE,subword_only,
                        output,exclude,rul);

      if (rul.left.subsyst != NULL) break;

      subsyst = subsyst->get_next_subsystem();
    }

    if (subsyst == NULL) continue;  // The loop ended normally.

    ll    = rul.left.length;
    lr    = rul.right.length;
    lout += lr - ll;

    if (lout > lout_max) {
      // No room to use rule.  Abort rewriting.

      while (bbot < lout_max)
        output.set_entry(lb++, output[bbot++]);
        output.set_length(lb);
      return;
    }

    n_applications++;
    total_left += ll;
    changed = TRUE;
    lb    -= ll;
    bbot  -= lr;
    rul.right.unpack(output,bbot);

    if (rewrite_verbose) {
      cout << "Which becomes:   ";
      for (i = 0; i < lb; i++) cout << output[i] << " ";
      for (i = bbot; i < lout_max; i++)
         cout << output[i] << " ";
      cout << endl;
    }
  }
  
  output.set_length(lb);
  reduced = TRUE;
}



/////////////////////////////////////////////////////////////////
//
//             rewriting_system::rewrite_global
//
/////////////////////////////////////////////////////////////////

void rewriting_system::rewrite_global(
  bool            subword_only,
  vector_long    &input,
  vector_long    &output,
  rule_reference  exclude,
  bool           &changed,
  bool           &reduced
)
{
  long         lin,lout,lout_max,lnew,i,j,off,ll,lr,diff,diff1;
  rule_type    r_type;
  packed_rule  rul,rul1;
  subsystem   *subsyst;

  // The max_length of output will not be changed.

  changed = FALSE;
  reduced = FALSE;

  lin = input.get_length();
  lout_max = output.get_max_length();

  if (lin > lout_max)
    error("output buffer too small in rewrite_global.");

  if (rewrite_verbose) {
    cout << "Global rewriting" << endl;
    input.print();
  }

  n_rewrite++;
  total_input += lin;

  if (n_sub <= 0) {
    if (rewrite_verbose) {
      cout << "No subsystems!" << endl;
    }

    output.set_copy(input);
    reduced = TRUE;
    return;
  }

  subsyst = first_subsystem;

  while (subsyst != NULL) {
    if (patterns && !use_patterns) {
      r_type = subsyst->get_r_type();

      if (r_type == SUBWORD_PATTERN || r_type == PREFIX_PATTERN) {
        subsyst = subsyst->get_next_subsystem();
        continue;
      }
    }

    subsyst->check_aux(REWRITE,subword_only,lout_max);
    subsyst = subsyst->get_next_subsystem();
  }

  lout = lin;

  for (i = 0; i < lin; i++) output.set_entry(i,input[i]);
  output.set_length(lout);

  for (;;) {
    subsyst = first_subsystem;
    rul.left.subsyst = NULL;
    diff = LONG_MAX;

    while (subsyst != NULL) {
      if (patterns && !use_patterns) {
        r_type = subsyst->get_r_type();

        if (r_type == SUBWORD_PATTERN || r_type == PREFIX_PATTERN) {
          subsyst = subsyst->get_next_subsystem();
          continue;
        }
      }

      rul1 = subsyst->global_test(subword_only,output,
                                       exclude);

      if (rul1.left.subsyst != NULL) {
        diff1 = rul1.right.length - rul1.left.length;

        if (rul.left.subsyst == NULL || diff1 < diff) {
          rul = rul1;
          diff = diff1;
        }
      }

      subsyst = subsyst->get_next_subsystem();
    }

    if (rul.left.subsyst == NULL) break;  // No left side was
                                          // found.

    ll   = rul.left.length;
    lr   = rul.right.length;
    off  = rul.offset;
    lnew = lout + lr - ll;

    if (lnew > lout_max) {
      // No room to use rule.  Abort rewriting.

      if (rewrite_verbose) {
        cout << "bsize exceeded" << endl;
        output.print();
      }

      return;
    }

    n_applications++;
    total_left += ll;
    changed = TRUE;

    if (ll > lr)
      for (i = off, j = i-ll+lr; i < lout; )
        output.set_entry(j++,output[i++]);

    else if (ll < lr)
      for (i = lout-1, j = i-ll+lr; i >= off; )
        output.set_entry(j--,output[i--]);

    rul.right.unpack(output,off-ll);

    lout = lnew;
    output.set_length(lout);

    if (rewrite_verbose) {
      cout << "Which becomes ";
      output.print();
    }
  }
  
  reduced = TRUE;
}



/////////////////////////////////////////////////////////////////
//
//            rewriting_system::reduce_all_subsystems
//
/////////////////////////////////////////////////////////////////

bool rewriting_system::reduce_all_subsystems(
  rule_type   rtype
)
{
  bool        new_call,new_loop,new_subsyst;
  subsystem  *subsyst;

  if (print_level >= 1) {
    cout << "Reducing all subsystems of type " << rtype << endl;

    if (logll > 0)
      log_stream <<
        "Reducing all subsystems of type " << rtype << endl << flush;
  }

  new_call = FALSE;

  do {
    new_loop = FALSE;
    subsyst = first_subsystem;

    while (subsyst != NULL) {
      if ( subsyst->get_r_type() == rtype) {
        new_subsyst = subsyst->reduce();
        new_loop    = new_loop||new_subsyst;
      }

      subsyst = subsyst->get_next_subsystem();
    }

    if (rtype == PREFIX_RULE && broken_prefix_rules != NULL) {
      new_loop = new_loop || broken_prefix_rules->reduce();

      if (0 == broken_prefix_rules->get_nrules() ) {
        delete broken_prefix_rules;
        broken_prefix_rules = NULL;
      }
    }

    else if (rtype == SUBWORD_RULE &&
             broken_subword_rules != NULL) {
      new_loop = new_loop || broken_subword_rules->reduce();

      if (0 == broken_subword_rules->get_nrules() ) {
        delete broken_subword_rules;
        broken_subword_rules = NULL;
      }
    }

    new_call = new_call||new_loop;
  } while (new_loop);

  return new_call;
}



/////////////////////////////////////////////////////////////////
//
//            rewriting_system::reduce_save_list
//
/////////////////////////////////////////////////////////////////

bool rewriting_system::reduce_save_list(void)
{
  bool        new_call,new_loop,new_subsyst;
  subsystem  *subsyst;

  if (print_level >= 1) {
    cout << "Reducing the save list." << endl;

    if (logll > 0)
      log_stream << "Reducing the save list." << endl << flush;
  }

  new_call = FALSE;

  do {
    new_loop = FALSE;
    subsyst = first_save;

    while (subsyst != NULL) {
      new_subsyst = subsyst->reduce();
      new_loop = new_loop || new_subsyst;
      subsyst = subsyst->get_next_save();
    }

    new_call = new_call || new_loop;
  } while (new_loop);

  return new_call;
}



/////////////////////////////////////////////////////////////////
//
//                rewriting_system::overlap
//
/////////////////////////////////////////////////////////////////

void rewriting_system::overlap(
  subsystem  *subsyst1,
  subsystem  *subsyst2,
  long        start_orig,
  long        minlen,
  long        maxlen,
  subsystem **avoid,
  long        navoid,
  long        maxlev,
  long       &neworig,
  long       &nnew,
  bool       &successful
)
{
  long                 free1,orig1,lr1,ll1,llmax2,i,lov,ll2,lr2;
  long                 ll3,lr3;
  long                 g,status,offset,nrules1,incr;
  bool                 changed,reduced;
  rule_type            rtp1,rtp2;
  vector_long_segment  useg;
  packed_rule          p_rule;
  rule                 rule1,rule2,rule3;
  rule_reference       exclude = {NULL, 0};

  update_time();

  if (print_level >= 1) {
    cout << "Forming overlaps of subsystem at "
         << (ulong)subsyst1 << " with subsystem at "
         << (ulong)subsyst2 << endl;

    if (logll > 0)
      log_stream << "Forming overlaps of subsystem at "
                 << (ulong)subsyst1 << " with subsystem at "
                 << (ulong)subsyst2 << endl << flush;
  }

  rtp1 = subsyst1->get_r_type();  
  rtp2 = subsyst2->get_r_type();

  successful = FALSE;
  nnew       = 0;
  neworig    = -1;

  if (PREFIX_PATTERN  == rtp1 ||
      PREFIX_PATTERN  == rtp2 ||
      SUBWORD_PATTERN == rtp1 ||
      SUBWORD_PATTERN == rtp2   ) {
    cout << "Overlaps with patterns not currently supported"
         << endl;

    return;
  }

  if (SUBWORD_RULE != rtp2) {
    cout <<
    "Second subsystem for overlaps must contain subword rules"
         << endl;

    return;
  }

  successful = TRUE;
  rule3.set_max_length(bsize);
  rule3.r_type = rtp1;
  llmax2 = subsyst2->get_llmax();

  // Check that subsystems listed in avoid have the auxiliary
  // space they need.

  for (i = 0; i < navoid; i++)
    avoid[i]->check_aux(OVERLAP,TRUE,maxlen);

  // Process the rules in subsyst1.

  nrules1 = subsyst1->get_nrules();
  incr    = max(10,nrules1/500);
  initialize_dots('.',incr,50);
  free1 = subsyst1->get_free();
  orig1 = start_orig;

  while (orig1 < free1) {
    print_dots();
    neworig = orig1;
    subsyst1->rule_by_origin(orig1,p_rule);
    if (p_rule.redundant) continue;

    ll1 = p_rule.left.length;
    lr1 = p_rule.right.length;

    if (ll1 >= maxlen || ll1+llmax2 <= minlen) continue;
    subsyst1->unpack(p_rule,rule1);

    if (maxlev < word_level(rule1.left) ) continue;

    for (i = 0; i < navoid; i++)
      if (avoid[i]->cummulative_test(rule1.left,rtp1))
        error("left side not reduced in overlap.");

    // Begin backtrack search for suffixes C which produce
    // minimal overlaps.

    lov = ll1;
    
    // make sure left side of rule1 is big enough for backtrack.

    rule1.left.set_max_length(maxlen);
    rule1.left.set_entry(lov++,0);
    rule1.left.set_length(lov);

    for (;;) {
      for (i = 0; i < navoid; i++) {
        status = avoid[i]->overlap_test(rule1.left,ll1,p_rule);

        // status == 1, rule1.left ends with rule returned in
        //              p_rule, whose left side has length
        //              greater than lov-ll1.
        // status == 2, rule1.left ends with rule returned in
        //              p_rule, whose left side has length less
        //              than or equal to lov-ll1.
        // status == 3, no suffix of rule1.left which is a prefix
        //              of a rule in subsystem avoid[i] has
        //              length more than lov-ll1.
        // status == 4, none of 1, 2, or 3.

        if (status == 1)
          if (avoid[i] == subsyst2) goto overlap_found;
          else goto increment;
          
        if (status == 2) goto increment;

        if (status == 3 && avoid[i] == subsyst2) goto increment;
      }

      if (lov >= maxlen) goto increment;
      rule1.left.set_entry(lov++,0);
      rule1.left.set_length(lov);
      continue;

     overlap_found:
      noverlaps++;
      if (lov < minlen) goto increment;
      offset = lov - p_rule.left.length;
      if (offset >= ll1) goto increment;
      subsyst2->unpack(p_rule,rule2);
      ll2 = rule2.left.get_length();
      lr2 = rule2.right.get_length();

      if (lr1 + lov - ll1 > bsize ||
          lr2 + offset    > bsize   ) {
        successful = FALSE;
        goto increment;
      }

      rule3.left.set_copy(rule1.right);
      useg.segment(rule1.left,ll1,lov-ll1);
      rule3.left.append(useg);
      useg.segment(rule1.left,0,offset);
      rule3.right.set_copy(useg);
      rule3.right.append(rule2.right);

      rewrite(rule3,exclude,changed,reduced);
      if ( orient_rule(rule3) ) goto increment;

      if (!reduced) {
         successful = FALSE;
         goto increment;
      }

      balance_rule(rule3);
      ll3 = rule3.left.get_length();
      lr3 = rule3.right.get_length();
      
      if (ll3 <= savell && lr3 <= savelr) {
        save_rule(rule3);
        nnew++;

        if (abort() ) goto done;
      }

      else successful = FALSE;

     increment:
      while(lov > 0 && (g = (rule1.left)[lov-1]) == t_gens-1)
        lov--;
      if (lov <= ll1) break;
      rule1.left.set_entry(lov-1,g+1);
      rule1.left.set_length(lov);
    }  // for (;;)
  }  // while (orig1 < free1)

  neworig = -1;

 done:
  end_dots();

  return;
}



/////////////////////////////////////////////////////////////////
//
//                rewriting_system::random_overlap
//
/////////////////////////////////////////////////////////////////

void rewriting_system::random_overlap(
  subsystem  *subsyst1,
  subsystem  *subsyst2,
  long        maxlen,
  subsystem **avoid,
  long        navoid,
  double     *count
)
{
  long                 free1,orig1,lr1,ll1,llmax2,i,lov,ll2,lr2;
  long                 ll3,lr3,j,ng;
  long                 g,status,offset,nrules1,incr;
  double               num;
  bool                 changed,reduced;
  rule_type            rtp1,rtp2;
  vector_long          gen_list;
  vector_long_segment  useg;
  packed_rule          p_rule;
  rule                 rule1,rule2,rule3;
  rule_reference       exclude = {NULL, 0};

  update_time();

  gen_list.set_max_length(t_gens);
  rtp1 = subsyst1->get_r_type();  
  rtp2 = subsyst2->get_r_type();

  if (PREFIX_PATTERN  == rtp1 ||
      PREFIX_PATTERN  == rtp2 ||
      SUBWORD_PATTERN == rtp1 ||
      SUBWORD_PATTERN == rtp2   ) {
    cout << "Overlaps with patterns not currently supported"
         << endl;

    return;
  }

  if (SUBWORD_RULE != rtp2) {
    cout <<
    "Second subsystem for overlaps must contain subword rules"
         << endl;

    return;
  }

  rule3.set_max_length(bsize);
  rule3.r_type = rtp1;
  llmax2 = subsyst2->get_llmax();

  // Check that subsystems listed in avoid have the auxiliary
  // space they need.

  for (i = 0; i < navoid; i++)
    avoid[i]->check_aux(OVERLAP,TRUE,maxlen);

  // Process random rules in subsyst1.

  nrules1 = subsyst1->get_nrules();
  subsyst1->rule_by_number(irand(0)%nrules1,p_rule);
  if (p_rule.redundant) return;

  ll1 = p_rule.left.length;
  lr1 = p_rule.right.length;

  if (ll1 >= maxlen) return;
  subsyst1->unpack(p_rule,rule1);

  for (i = 0; i < navoid; i++)
    if (avoid[i]->cummulative_test(rule1.left,rtp1))
      error("left side not reduced in overlap.");

  // Begin backtrack search for suffixes C which produce
  // minimal overlaps.

  lov = ll1;
  num = (double) nrules1;
  rule1.left.set_max_length(maxlen);

  while (lov < maxlen) {
    lov++;
    ng = 0;
    rule1.left.set_length(lov);

    for (j = 0; j < t_gens; j++) {
      rule1.left.set_entry(lov-1,j);

      for (i = 0; i < navoid; i++) {
        status = avoid[i]->overlap_test(rule1.left,ll1,p_rule);

        // status == 1, rule1.left ends with rule returned in
        //              p_rule, whose left side has length
        //              greater than lov-ll1.
        // status == 2, rule1.left ends with rule return in
        //              p_rule, whose left side has length less
        //              than or equal to lov-ll1.
        // status == 3, no suffix of rule1.left which is a prefix
        //              of a rule in subsystem avoid[i] has
        //              length more than lov-ll1.
        // status == 4, none of 1, 2, or 3.

        if (status == 1)
          if (avoid[i] == subsyst2) {
            gen_list.set_entry(ng++,j);
            break;
          }
          else break;
          
        if (status == 2) break;

        if (status == 3 && avoid[i] == subsyst2) break;
      }

      if (i >= navoid) gen_list.set_entry(ng++,j);
    }

    if (0 == ng) return;

    num *= ng;
    count[lov-ll1+1] += num;
    rule1.left.set_entry(lov-1,gen_list[irand(0)%ng]);
    status = subsyst2->overlap_test(rule1.left,ll1,p_rule);
    if (1 == status) goto overlap_found;
  }

  if (status != 1) return;

 overlap_found:
  noverlaps++;
  offset = lov - p_rule.left.length;
  if (offset >= ll1) return;
  subsyst2->unpack(p_rule,rule2);
  ll2 = rule2.left.get_length();
  lr2 = rule2.right.get_length();

  if (lr1 + lov - ll1 > bsize ||
      lr2 + offset    > bsize   ) return;

  rule3.left.set_copy(rule1.right);
  useg.segment(rule1.left,ll1,lov-ll1);
  rule3.left.append(useg);
  useg.segment(rule1.left,0,offset);
  rule3.right.set_copy(useg);
  rule3.right.append(rule2.right);

  rewrite(rule3,exclude,changed,reduced);
  if ( orient_rule(rule3) ) return;

  if (!reduced) return;

  balance_rule(rule3);
  ll3 = rule3.left.get_length();
  lr3 = rule3.right.get_length();
  if (ll3 <= savell && lr3 <= savelr) save_rule(rule3);

  return;
}



/////////////////////////////////////////////////////////////////
//
//               rewriting_system::orient_rule
//
/////////////////////////////////////////////////////////////////

bool rewriting_system::orient_rule(
  rule  &rul
)
{
  long  cmp;

  cmp = comp(rul.left, rul.right);
  
//  cmp = (this->*compare)(rul.left, rul.right);

  if (cmp == 0) return TRUE;
  if (cmp == -1) rul.switch_sides();
  return FALSE;
}



/////////////////////////////////////////////////////////////////
//
//                rewriting_system::balance_rule
//
/////////////////////////////////////////////////////////////////

void rewriting_system::balance_rule(
  rule     &rul
)
{
  long                ll,lr,lrmax,i,g,linv,ls,lp,lb,j;
  long                sgn;
  bool                subword_only,changed,reduced;
  vector_long_segment lseg;
  rule_reference      exclude = {NULL,0};

  // We assume the left and right sides are distinct, that the
  // rule is oriented correctly, and that both sides are reduced.

  if (PREFIX_PATTERN  == rul.r_type ||
      SUBWORD_PATTERN == rul.r_type   ) {
    // Currently no balancing is done for patterns.

    return;
  }

  nbalance++;

  if (rul.r_type == SUBWORD_RULE) subword_only = TRUE;
  else subword_only == FALSE;

  ll    = rul.left.get_length();
  lr    = rul.right.get_length();
  lrmax = rul.right.get_max_length();
  balance_vector.set_max_length(lrmax);

 loop:
  if (rul.r_type == SUBWORD_RULE) {
    // Cancel any common cancelable generators from the front.

    i = 0;

    while (   i < ll && 
              i < lr &&
           (g = (rul.left)[i]) == (rul.right)[i] &&
            gens[g].get_cancelable()                 ) i++;

    if (i > 0) {
      rul.left.truncate(i);
      rul.right.truncate(i);
      ll -= i;
      lr -= i;
    }
  }

  // Cancel common cancelable generators from the end.

  while ( ll > 0 &&
          lr > 0 &&
         (g = rul.left[ll-1]) == rul.right[lr-1] &&
         gens[g].get_cancelable()                   ) {
    ll--;
    lr--;
  }

  rul.left.set_length(ll);
  rul.right.set_length(lr);

  // Find the length of the longest obviously invertible suffix
  // of the left side (if any) and move its inverse to the end
  // of the right side.

  g = (rul.left)[ll-1];

  if (gens[g].get_invertible() ) {
    linv = gens[g].get_inv_length();
    ls = 1;

    while (ls < ll &&
           ls < linv &&
           (rul.left)[ll-ls-1] ==
                gens[g].get_inv_term(linv-ls) ) ls++;

    lb = lr + linv + 1 - ls;

    if (lb <= lrmax) {
      balance_vector.set_copy(rul.right);
      balance_vector.set_length(lb);

      for (j = 0; j < linv+1-ls; j++)
        balance_vector.set_entry(lr+j, gens[g].get_inv_term(j) );

      rewrite_from_left(
        subword_only,balance_vector,balance_vector,exclude,
        changed,reduced);

      if (!reduced) goto left_end;

      lseg.segment(rul.left,0,ll-ls);

      sgn = comp(lseg,balance_vector);
      
//      sgn = (this->*compare)(lseg,balance_vector);

      if (sgn > 0) {
        ll -= ls;
        rul.left.set_length(ll);
        rul.right.set_copy(balance_vector);
        lr = rul.right.get_length();
        goto loop;
      }
    }
  }

 left_end:
  if (rul.r_type == SUBWORD_RULE) {
    // Find the longest obviously invertible prefix of the left
    // side (if any) and move it to the front of the right side.

    g = (rul.left)[0];

    if (gens[g].get_invertible() ) {
      linv = gens[g].get_inv_length();
      lp = 1;

      while (lp < ll &&
             lp < linv &&
             (rul.left)[lp] ==
                  gens[g].get_inv_term(lp-1) ) lp++;

      lb = lr + linv + 1 - lp;

      if (lb <= lrmax) {
        for (j = 0; j < linv+1-lp; j++)
          balance_vector.set_entry(j,gens[g].get_inv_term(j+lp-1) );
        balance_vector.set_length(linv+1-lp);
        balance_vector.append(rul.right);

        rewrite_from_left(subword_only,balance_vector,
                        balance_vector,exclude,changed,reduced);
        if (!reduced) goto no_more;

        lseg.segment(rul.left,lp,ll-lp);

        sgn = comp(lseg,balance_vector);

//        sgn = (this->*compare)(lseg,balance_vector);

        if (sgn > 0) {
          rul.left.truncate(lp);
          ll -= lp;
          rul.right.set_copy(balance_vector);
          lr = rul.right.get_length();
          goto loop;
        }
      }
    }
  }

//  cout << " output " << endl;
//  rul.print();
//  cout << endl;

 no_more:
  return;
}



/////////////////////////////////////////////////////////////////
//
//             rewriting_system::heuristics (one rule)
//
/////////////////////////////////////////////////////////////////

void rewriting_system::heuristics(
  rule   &rul
)
{
  long                ll,lr,lrmax,i,g,linv,ls,lp,lb,j;
  long                sgn;
  bool                subword_only,changed,reduced;
  vector_long_segment lseg;
  rule_reference      exclude = {NULL,0};

  if (PREFIX_PATTERN  == rul.r_type ||
      SUBWORD_PATTERN == rul.r_type   ) {
    // Heuristics are not applied to patterns.

    return;
  }

  // We assume the left and right sides are distinct and that
  // the rule is oriented correctly and balanced.

  heuristic_rule.set_max_length(bsize);
  heuristic_rule.r_type = rul.r_type;

  if (rul.r_type == SUBWORD_RULE) subword_only = TRUE;
  else subword_only = FALSE;

  ll    = rul.left.get_length();
  lr    = rul.right.get_length();
  lrmax = rul.right.get_max_length();
  heuristic_vector.set_max_length(lrmax);

  // Find the length of the longest obviously invertible suffix
  // of the left side (if any) and move its inverse to the end
  // of the right side.

  g = (rul.left)[ll-1];

  if (gens[g].get_invertible() ) {
    linv = gens[g].get_inv_length();
    ls = 1;

    while (ls < ll &&
           ls < linv &&
           (rul.left)[ll-ls-1] ==
                gens[g].get_inv_term(linv-ls) ) ls++;

    lb = lr + linv + 1 - ls;

    if (lb <= lrmax) {
      heuristic_vector.set_copy(rul.right);
      heuristic_vector.set_length(lb);

      for (j = 0; j < linv+1-ls; j++)
        heuristic_vector.set_entry(lr+j,gens[g].get_inv_term(j));

      rewrite_from_left(subword_only,heuristic_vector,
                       heuristic_vector,exclude,changed,reduced);
      if (!reduced) goto left_end;

      lseg.segment(rul.left,0,ll-ls);

      sgn = comp(lseg,heuristic_vector);

//      sgn = (this->*compare)(lseg,heuristic_vector);

      if (sgn < 0) {
        heuristic_rule.right.set_copy(rul.left);
        heuristic_rule.right.set_length(ll-ls);
        heuristic_rule.left.set_copy(heuristic_vector);
        balance_rule(heuristic_rule);

        if (heuristic_rule.left.get_length() <= savell &&
            heuristic_rule.right.get_length() <= savelr   )
          save_rule(heuristic_rule);
      }
    }
  }

 left_end:
  if (rul.r_type == SUBWORD_RULE) {
    // Find the longest obviously invertible prefix of the left
    // side (if any) and move it to the front of the right side.

    g = (rul.left)[0];

    if (gens[g].get_invertible() ) {
      linv = gens[g].get_inv_length();
      lp = 1;

      while (lp < ll &&
             lp < linv &&
             (rul.left)[lp] ==
                  gens[g].get_inv_term(lp-1) ) lp++;

      lb = lr + linv + 1 - lp;

      if (lb <= lrmax) {
        for (j = 0; j < linv+1-lp; j++)
          heuristic_vector.set_entry(j,gens[g].get_inv_term(j+lp-1) );
        heuristic_vector.set_length(linv+1-lp);
        heuristic_vector.append(rul.right);

        rewrite_from_left(subword_only,heuristic_vector,
                      heuristic_vector,exclude,changed,reduced);
        if (!reduced) goto no_more;

        lseg.segment(rul.left,lp,ll-lp);

        sgn = comp(lseg,heuristic_vector);

//        sgn = (this->*compare)(lseg,heuristic_vector);

        if (sgn < 0) {
          heuristic_rule.right.set_copy(rul.left);
          heuristic_rule.right.truncate(lp);
          heuristic_rule.left.set_copy(heuristic_vector);
          balance_rule(heuristic_rule);

          if (heuristic_rule.left.get_length() <= savell &&
              heuristic_rule.right.get_length() <= savelr   )
            save_rule(heuristic_rule);
        }
      }
    }
  }

 no_more:
  return;
}



/////////////////////////////////////////////////////////////////
//
//         rewriting_system::heuristics (save_list)
//
/////////////////////////////////////////////////////////////////

void rewriting_system::heuristics(void)
{
  int          more;
  subsystem   *subsyst;

  if (print_level >= 1) {
    cout << "Applying heuristics to entire save list." << endl;

    if (logll > 0)
      log_stream << "Appyling heuristics to entire save list."
                 << endl << flush;
  }

  more = 1;

  while (more) {
    more = 0;
    subsyst = first_save;

    while (subsyst != NULL) {
      more = more || subsyst->heuristics();
      if (abort() ) return;
      subsyst = subsyst->get_next_save();
    }
  }

  return;
}



/////////////////////////////////////////////////////////////////
//
//           rewriting_system::initialize_save_list
//
/////////////////////////////////////////////////////////////////

void rewriting_system::initialize_save_list(
  rule_type  rtp,
  long       enm,    // Estimated number of rules.
  long       ell     // Estimated length of left sides.
)
{
  long        minll,tlen,dim;
  subsystem  *subsyst;

  minll = max( (3*ell)/4, ell-2);
  minll = max(minll,1);

  subsyst = new subsystem(this,rtp,RABIN_KARP);
  subsyst->make_rabin_karp(enm,minll);

  nsave = 0;
  first_save = last_save = subsyst;
  add_subsystem(subsyst);
}



/////////////////////////////////////////////////////////////////
//
//              rewriting_system::append_save_list
//
/////////////////////////////////////////////////////////////////

void rewriting_system::append_save_list(
  subsystem  *subsyst
)
{
  subsystem   *old_subs, *subs = first_save;

//  subs = first_save;

  while (subs != NULL) {
    subsyst->append(subs);
    old_subs = subs;
    subs     = subs->get_next_save();
    remove_subsystem(old_subs);
  }
}



/////////////////////////////////////////////////////////////////
//
//               rewriting_system::remove_subsystem
//
/////////////////////////////////////////////////////////////////

void rewriting_system::remove_subsystem(
  subsystem   *subsyst
)
{
  subsystem  *nsubs,*psubs,*nsave,*psave;

  nsubs = subsyst->get_next_subsystem();
  psubs = subsyst->get_prev_subsystem();
  nsave = subsyst->get_next_save();
  psave = subsyst->get_prev_save();

  if (first_subsystem == subsyst) first_subsystem = nsubs;
  if (last_subsystem  == subsyst) last_subsystem  = psubs;
  if (first_save      == subsyst) first_save      = nsave;
  if (last_save       == subsyst) last_save       = psave;

  if (nsubs != NULL) nsubs->set_prev_subsystem(psubs);
  if (psubs != NULL) psubs->set_next_subsystem(nsubs);
  if (nsave != NULL) nsave->set_prev_save(psave);
  if (psave != NULL) psave->set_next_save(nsave);

  n_sub--;

  delete subsyst;
}



/////////////////////////////////////////////////////////////////
//
//                rewriting_system::save_rule
//
/////////////////////////////////////////////////////////////////

void rewriting_system::save_rule(
  rule    &rul
)
{
  long        ll,win,pwin,dim;
  rule_type   rtype;
  subsystem  *subsyst,*prev,*subsyst1;
  
  nsave++;
  nsave_total++;

  subsyst = first_save;
  if (subsyst == NULL)
    error("save list not initialized.");

  ll = rul.left.get_length();

  if (ll <= logll) {
    rtype = subsyst->get_r_type();

    if      (rtype == SUBWORD_RULE)    log_stream << "SR ";
    else if (rtype == PREFIX_RULE)     log_stream << "PR ";
    else if (rtype == SUBWORD_PATTERN) log_stream << "SP ";
    else if (rtype == PREFIX_PATTERN)  log_stream << "PP ";

    print_word(log_stream,logform,rul.left);
    log_stream << " ";
    print_word(log_stream,logform,rul.right);
    log_stream << endl << flush;
  }

  prev = NULL;

  while (subsyst != NULL) {
    win = subsyst->get_window();
    if (win <= ll) break;
    prev = subsyst;
    subsyst = subsyst->get_next_save();
  }

  if (prev == NULL) {
    subsyst->add_rule(rul,TRUE,FALSE);
    return;
  }

  pwin = prev->get_window();

  while(pwin > ll) pwin = (4*pwin)/5;

  if (subsyst != NULL && pwin <= win) {
    subsyst->add_rule(rul,TRUE,FALSE);
    return;
  }

  subsyst1 = new subsystem(this,rul.r_type,RABIN_KARP);
  subsyst1->make_rabin_karp(1000,pwin);
  prev->set_next_save(subsyst1);
  subsyst1->set_prev_save(prev);
  subsyst1->set_next_save(subsyst);

  if (subsyst != NULL) subsyst->set_prev_save(subsyst1);
  else last_save = subsyst1;

  add_subsystem(subsyst1);
  subsyst1->add_rule(rul,TRUE,FALSE);
}



/////////////////////////////////////////////////////////////////
//
//                   rewriting_system::kb_basic
//
/////////////////////////////////////////////////////////////////

//  Keep just one primary subsystem of each type.  Return TRUE if
//  there is a change and FALSE otherwise.

bool rewriting_system::kb_basic(
  long      maxlev,      // the maximum level of the left part of
                         // an overlap
  long      maxlen,      // the maximum length of overlaps 
  long     &num_iter,    // the maximum number of calls to
                         // overlap allowed.  a value of -1
                         // means infinitely many calls are
                         // permitted.
  long      prevlen,     // all overlaps of length up to this
                         // value have been processed.
  long      flag         // see kb_two_step.
)
{
  rule_type    type0;
  bool         successful,confluent;
  long         nrules0,pass_number;
  long         nrules1,est_num,est_ll,est_lr;
  long         maxllnew,start_orig,minlen;
  long         navoid,nnew,neworig,change0,change1;
  long         first_pass;
  subsystem  *subsyst0,*subsyst1,*avoid[1];
  
  if (print_level >= 1) {
    cout << "\nkb_basic:  maxlen = " << maxlen 
         << ", num_iter = " << num_iter << endl << endl;

    if (logll > 0)
      log_stream << "\nkb_basic:  maxlen = " << maxlen 
                  << ", num_iter = " << num_iter << endl << flush;
  }

  // Set parameters used in reduce_all_subsystems.

  // There should be at most two subsystems and if there are two,
  // the first one should be a subword system and the second one
  // should a prefix system.

  confluent = TRUE;

  if (n_sub    == 0 ||
      num_iter == 0   ) return FALSE;

  if (n_sub > 2) {
    cout << "too many subsystems in kb_basic" << endl;
    exit(1);
  }

  subsyst0   = first_subsystem;
  type0      = subsyst0->get_r_type();
  nrules0    = subsyst0->get_nrules();

  if (type0 == PREFIX_RULE) {
    if (n_sub > 1) goto invalid_type;
    if (nrules0 == 0) return TRUE;

    // In the case of prefix rules only, forming overlaps is the
    // same as mutual reduction.

    est_num = max(100,nrules0/2);

    initialize_save_list(PREFIX_RULE,est_num,savell);
    (void) reduce_all_subsystems(PREFIX_RULE);
    if (broken_prefix_rules != NULL) confluent = FALSE;
    append_save_list(subsyst0);
    if (subsyst0->origin_in_use() ) subsyst0->make_origin();
    subsyst0->make_index();

    return confluent;
  }

  // Form overlaps of subword rules.

  pass_number = 0;
  first_pass = TRUE;
  if (num_iter == 0) return FALSE;

  do {
    pass_number++;

    if (print_level >= 2)
      cout << "Pass number " << pass_number
           << " on subword rules" << endl;

    change0 = FALSE;
    nrules0 = subsyst0->get_nrules();

    if (print_level >= 2)
      cout << "There are " << nrules0 << " rules" << endl;

    initialize_save_list(SUBWORD_RULE,max(1000,nrules0/2),savell);

    start_orig = 0;

    if (first_pass) {
      minlen     = prevlen+1;
      first_pass = FALSE;
    }

    else minlen = 0;

    navoid     = 1;
    avoid[0]   = subsyst0;

    overlap(subsyst0,subsyst0,start_orig,minlen,maxlen,
            avoid,navoid,maxlev,neworig,nnew,successful);

    confluent = confluent && successful;

    if (nnew > 0) {
      change0 = TRUE;
    }

    (void) reduce_all_subsystems(SUBWORD_RULE);
    if (broken_subword_rules != NULL) confluent = FALSE;
    append_save_list(subsyst0);

    if (subsyst0->origin_in_use() ) subsyst0->make_origin();
    subsyst0->make_index();
    if (num_iter > 0) num_iter--;
  }
  while (change0 && num_iter != 0);

  if (change0) confluent = FALSE;
  if (n_sub == 1) return confluent;

  subsyst1   = subsyst0->get_next_subsystem();
  nrules1    = subsyst1->get_nrules();

  if (subsyst1->get_r_type() != PREFIX_RULE) goto invalid_type;
  if (nrules1 == 0) return confluent;

  // Since new subword rules may have been found, the prefix
  // rules may not be reduced.

  // Form overlaps of prefix rules with subword rules.

  pass_number = 0;

  do {
    pass_number++;

    if (print_level >= 2)
      cout << "Pass number " << pass_number << " on prefix rules"
           << endl;

    change1 = FALSE;
    nrules1 = subsyst1->get_nrules();

    if (print_level >= 2)
      cout << "There are " << nrules1 << " rules" << endl;

    initialize_save_list(PREFIX_RULE,max(1000,nrules1/2),savell);

    start_orig = 0;

    // Too many things may have happened to assume that
    // prevlen is still valid.

    minlen     = 0;
    navoid     = 1;
    avoid[0]   = subsyst0;

    overlap(subsyst1,subsyst0,start_orig,minlen,maxlen,
            avoid,navoid,maxlev,neworig,nnew,successful);

    confluent = confluent && successful;
    if (num_iter > 0) num_iter--;
    if (nnew > 0) change1 = TRUE;

    // Mutually reducing all prefix subsystems includes forming
    // all overlaps between the prefix rules.

    (void) reduce_all_subsystems(PREFIX_RULE);
    if (broken_prefix_rules != NULL) confluent = FALSE;
    append_save_list(subsyst1);
    subsyst1->make_index();
  }
  while (change1 && num_iter != 0);

  return confluent;

 invalid_type:
  error("invalid subsystem type in kb_basic");
  return confluent;
}


/////////////////////////////////////////////////////////////////
//
//                   system::kb_two_step
//
/////////////////////////////////////////////////////////////////

/*
    Keep two primary subsystems of each type.  Return TRUE
    if there is a change and FALSE otherwise.
*/

bool rewriting_system::kb_two_step(
  long      maxlev,      // the maximum level of left part of
                         // the overlap
  long      maxlen,      // the maximum length of overlaps
  long     &num_iter,    // the maximum number of basic steps
                         // allowed.  a value of -1
                         // means infinitely many calls are
                         // permitted.
  long      prevlen,     // all overlaps of length up to this
                         // value have been processed.
  long      flag         // the lowest order bit tells whether
                         // subword overlaps should be formed.
                         // the next lowest controls prefix rule
                         // overlaps.
)
{
  long         est_num,dim;
  long         next_minlen,minlen,navoid,nrules0,nrules1,nrules2;
  long         nnew,nnew01,nnew10,nnew11,pass_number;
  long         start_orig,neworig,maxllnew;
  long         subword_overlaps,prefix_overlaps;
  bool         change0,successful,confluent;
  rule_type    type,type0,type1;
  index_type   itype0,itype1;
  subsystem   *avoid[2],*subsyst;
  subsystem   *prefix0,*prefix1,*subword0,*subword1;
  subsystem   *prefix_pattern0,*subword_pattern0;
  
  nsave_total = 0;
  subword_overlaps = flag&1;
  prefix_overlaps  = (flag>>1)&1;

  if (print_level >= 1) {
    cout << "\nkb_two_step:  maxlen = " << maxlen 
         << ", num_iter = " << num_iter << endl;

    if (logll > 0)
      log_stream << "\nkb_two_step:  maxlen = " << maxlen 
                 << ", num_iter = " << num_iter << endl << flush;
  }

  if (n_sub == 0 || num_iter   == 0 ) return FALSE;

  prefix0 = subword0 = prefix_pattern0 = subword_pattern0 = NULL;

  subsyst  = first_subsystem;

  while (subsyst != NULL) {
    type = subsyst->get_r_type();

    if (type == PREFIX_RULE && prefix0 == NULL)
      prefix0 = subsyst;
    else if (type == SUBWORD_RULE && subword0 == NULL)
      subword0 = subsyst;
    else if (type == PREFIX_PATTERN && prefix_pattern0 == NULL)
      prefix_pattern0 = subsyst;
    else if (type == SUBWORD_PATTERN && subword_pattern0 == NULL)
      subword_pattern0 = subsyst;

    subsyst = subsyst->get_next_subsystem();
  }

  confluent = TRUE;
  next_minlen = prevlen + 1;

  // Form overlaps of subword rules, if any.

  if (!subword_overlaps || subword0 == NULL) goto prefix;

 subword:
  nrules0 = subword0->get_nrules();
  itype0  = subword0->get_i_type();

  if (print_level >= 2)
    cout << "First pass on subword rules" << endl
         << "There are " << nrules0 << " rules." << endl;

  if (num_iter == 0) return FALSE;

  change0 = FALSE;

  initialize_save_list(SUBWORD_RULE,max(1000,nrules0/2),savell);

  start_orig = 0;
  minlen = next_minlen;
  next_minlen = 0;
  navoid     = 1;
  avoid[0]   = subword0;

  overlap(subword0,subword0,start_orig,minlen,maxlen,
          avoid,navoid,maxlev,neworig,nnew,successful);

  confluent = confluent && successful;

  if (broken_subword_rules != NULL) {
    (void) broken_subword_rules->reduce();

    if (0 == broken_subword_rules->get_nrules() ) {
      delete broken_subword_rules;
      broken_subword_rules = NULL;
    }
  }

  if (num_iter > 0) num_iter--;

  if (nsave > 0) {
    change0 = TRUE;

    if (neworig != -1) {
      // processing of overlaps aborted

      (void) reduce_all_subsystems(SUBWORD_RULE);
      append_save_list(subword0);
      if (subword0->origin_in_use() ) subword0->make_origin();
      subword0->make_index();
      return FALSE;
    }

    if (use_heuristics) heuristics();
    (void) reduce_save_list();
    if (broken_subword_rules != NULL) confluent = FALSE;
    subword1 = new subsystem(this,SUBWORD_RULE,itype0);
                
    add_subsystem(subword1);
    append_save_list(subword1);

    if (subword1->origin_in_use() ) subword1->make_origin();
    subword1->make_index();
  }

  else {
    append_save_list(subword0);
  }

  pass_number = 1;

  while (change0) {
    nrules0 = subword0->get_nrules();
    nrules1 = subword1->get_nrules();

    if (abort() || num_iter == 0) {
      initialize_save_list(SUBWORD_RULE,max(1000,nrules1),savell);
      (void) reduce_all_subsystems(SUBWORD_RULE);
      subword0->append(subword1);
      remove_subsystem(subword1);
      append_save_list(subword0);

      if (subword0->origin_in_use() ) subword0->make_origin();
      subword0->make_index();
      return FALSE;
    }

    change0 = FALSE;
    pass_number++;

    if (print_level >= 2)
      cout << "Pass number " << pass_number
           << " on subword rules"<< endl;

    initialize_save_list(SUBWORD_RULE,max(1000,nrules1),savell);
    (void) subword0->reduce();
    nrules0 = subword0->get_nrules();

    if (print_level >= 2)
      cout << "There are " << nrules0 << " old rules and "
           << nrules1 << " new ones." << endl;

    if (nrules0 == 0 && nrules1 == 0) {
      printf("Where have all the rules gone in twostep?\n");
      exit(1);
    }

    start_orig = 0;
    minlen     = 0;
    avoid[0]   = subword0;
    avoid[1]   = subword1;
    navoid     = 2;

    nnew01 = nnew10 = nnew11 = 0;

    overlap(subword0,subword1,start_orig,minlen,maxlen,
            avoid,navoid,maxlev,neworig,nnew01,successful);

    confluent = confluent && successful;

    if (neworig == -1) {
      overlap(subword1,subword0,start_orig,minlen,maxlen,
              avoid,navoid,maxlev,neworig,nnew10,successful);

      confluent = confluent && successful;
    }

    if (neworig == -1) {
      overlap(subword1,subword1,start_orig,minlen,maxlen,
              avoid,navoid,maxlev,neworig,nnew11,successful);

      confluent = confluent && successful;
    }

    if (num_iter > 0) num_iter--;

    if (broken_subword_rules != NULL) {
      (void) broken_subword_rules->reduce();

      if (0 == broken_subword_rules->get_nrules() ) {
        delete broken_subword_rules;
        broken_subword_rules = NULL;
      }
    }

    if (nsave > 0) {
      change0 = TRUE;

      if (neworig == -1) {
        if (use_heuristics) heuristics();
        (void) reduce_save_list();
        if (broken_subword_rules != NULL) confluent = FALSE;

        subword0->append(subword1);
        remove_subsystem(subword1);
        subword1 = new subsystem(this,SUBWORD_RULE,itype0);

        add_subsystem(subword1);  
        append_save_list(subword1);

        if (subword0->origin_in_use() ) {
          subword0->make_origin();
          subword1->make_origin();
        }

        subword0->make_index();
        subword1->make_index();
      }

      else {
        (void) reduce_all_subsystems(SUBWORD_RULE);
        if (broken_subword_rules != NULL) confluent = FALSE;
        subword0->append(subword1);
        remove_subsystem(subword1);
        append_save_list(subword0);

        if (subword0->origin_in_use() )
          subword0->make_origin();

        subword0->make_index();
        return FALSE;
      }
    }

    else {
      append_save_list(subword1);
      subword0->append(subword1);
      remove_subsystem(subword1);
      if (subword0->origin_in_use() ) subword0->make_origin();
      subword0->make_index();
    }
  }

  // Now consider prefix rules.

 prefix:
  if (!prefix_overlaps || prefix0 == NULL) return confluent;

  nrules0 = prefix0->get_nrules();
  itype0  = prefix0->get_i_type();

  if (print_level >= 2)
    cout << endl << "Reducing old prefix rules" << endl;

  // Changes in subword rules may have made some prefix rules
  // reducible.

  est_num = max(100,nrules0/2);
  initialize_save_list(PREFIX_RULE,est_num,savell);
  (void) reduce_all_subsystems(PREFIX_RULE);
  if (broken_prefix_rules != NULL) confluent = FALSE;
  append_save_list(prefix0);
  prefix0->make_index();

  // Form overlaps of prefix rules with subword rules.

  nsave_total = 0;

 prefix_loop:
  if (abort() || num_iter == 0) return FALSE;
  if (subword0 == NULL) return TRUE;

  if (print_level >= 2)
    cout << "First pass on prefix rules" << endl;

  pass_number = 1;
  nrules0 = prefix0->get_nrules();

  if (nrules0 == 0) return confluent;

  if (print_level >= 2)
    cout << "There are " << nrules0 << " prefix rules" << endl;

  change0 = FALSE;

  initialize_save_list(PREFIX_RULE,max(1000,nrules0/2),savell);

  start_orig = 0;
  minlen     = 0;
  navoid     = 1;
  avoid[0]   = subword0;

  overlap(prefix0,subword0,start_orig,minlen,maxlen,
          avoid,navoid,maxlen,neworig,nnew,successful);

  confluent = confluent && successful;

  if (num_iter > 0) num_iter--;

  if (nnew > 0) {
    change0 = TRUE;
    (void) reduce_all_subsystems(PREFIX_RULE);
     if (broken_prefix_rules != NULL) confluent = FALSE;

    if (neworig != -1) {
      // processing of overlaps aborted

      append_save_list(prefix0);
      prefix0->make_index();
      return FALSE;
    }

    prefix1 = new subsystem(this,PREFIX_RULE,itype0);

    add_subsystem(prefix1);
    append_save_list(prefix1);

    prefix1->make_index();
  }

  else {
    append_save_list(prefix0);  // This just clears save list.
    return confluent;
  }

  // Make additional passes.

  while (change0) {
    if (abort() || num_iter == 0) {
      prefix0->append(prefix1);
      remove_subsystem(prefix1);
      prefix0->make_index();
      return FALSE;
    }

    pass_number++;

    if (print_level >= 2)
      cout << "Pass number " << pass_number
           << " on prefix rules" << endl;

    change0 = FALSE;
    nrules0 = prefix0->get_nrules();
    nrules1 = prefix1->get_nrules();

    if (nrules1 == 0) break;

    if (print_level >= 2)
      cout << "There are " << nrules0 << " old prefix rules and "
           << nrules1 << " new ones" << endl;

    initialize_save_list(PREFIX_RULE,max(1000,nrules1),savell);

    start_orig = 0;
    minlen     = 0;
    avoid[0]   = subword0;
    navoid = 1;

    overlap(prefix1,subword0,start_orig,minlen,maxlen,
            avoid,navoid,maxlev,neworig,nnew,successful);

    confluent = confluent && successful;
    if (num_iter > 0) num_iter--;

    if (nnew > 0) {
      change0 = TRUE;
      (void) reduce_all_subsystems(PREFIX_RULE);
      if (broken_prefix_rules != NULL) confluent = FALSE;
      prefix0->append(prefix1);
      remove_subsystem(prefix1);
      prefix1 = new subsystem(this,PREFIX_RULE,itype0);
                  
      add_subsystem(prefix1);
      append_save_list(prefix1);

      if (neworig == -1) {
        prefix0->make_index();
        prefix1->make_index();
      }

      else {
        prefix0->append(prefix1);
        remove_subsystem(prefix1);
        append_save_list(prefix0);
        prefix0->make_index();
        return FALSE;
      }
    }

    else {
      prefix0->append(prefix1);
      remove_subsystem(prefix1);
      append_save_list(prefix0);  // This just clears save list.
      prefix0->make_index();
    }
  }

 return confluent;

 invalid_type:
  printf("invalid subsystem type in twostep\n");
  exit(1);
}



/////////////////////////////////////////////////////////////////
//
//           rewriting_system::kb_random_two_step
//
/////////////////////////////////////////////////////////////////

bool rewriting_system::kb_random_two_step(
  long      maxlen,      // the maximum length of overlaps
  rule_type rtype,       // the type of AB in overlaps ABC
  long      ntrials      // the number of trials
)
{
  long         est_num,dim;
  long         next_minlen,minlen,navoid,nrules0,nrules1,nrules2;
  long         nnew,nnew01,nnew10,nnew11,pass_number;
  long         start_orig,neworig,maxllnew,ii,j;
  double      *count,num;
  bool         change0,successful,confluent,value;
  rule_type    type,type0,type1;
  index_type   itype0,itype1;
  subsystem   *avoid[2],*subsyst;
  subsystem   *prefix0,*prefix1,*subword0,*subword1;
  subsystem   *prefix_pattern0,*subword_pattern0;
  
  nsave_total = 0;

  if (print_level >= 1) {
    cout << "\nkb_random_two_step:  maxlen = " << maxlen 
         << ", ntrials = " << ntrials << endl;

    if (SUBWORD_RULE == rtype) cout << "Subword rules" << endl;
    else cout << "Prefix rules" << endl;

    if (logll > 0)
      log_stream << "kb_random_two_step:  maxlen = " << maxlen
         << ",ntrials = " << ntrials << endl << flush;
  }

  if (n_sub == 0) return FALSE;

  count = (double *) fxmllc( (maxlen+1)*sizeof(double) );

  if (NULL == count) {
    cout << "Unable to allocate space for counts" << endl;
    return FALSE;
  }

  for (j = 0; j <= maxlen; j++) count[j] = 0.0;

  prefix0 = subword0 = prefix_pattern0 = subword_pattern0 = NULL;

  subsyst = first_subsystem;

  while (subsyst != NULL) {
    type = subsyst->get_r_type();

    if (type == PREFIX_RULE && prefix0 == NULL)
      prefix0 = subsyst;
    else if (type == SUBWORD_RULE && subword0 == NULL)
      subword0 = subsyst;
    else if (type == PREFIX_PATTERN && prefix_pattern0 == NULL)
      prefix_pattern0 = subsyst;
    else if (type == SUBWORD_PATTERN && subword_pattern0 == NULL)
      subword_pattern0 = subsyst;

    subsyst = subsyst->get_next_subsystem();
  }

  if (SUBWORD_RULE == rtype) {
    // Form overlaps of subword rules, if any.

    if (subword0 == NULL) {
      cout << "There are no subword rules." << endl;
      value = FALSE;
      goto done;
    }

    nrules0 = subword0->get_nrules();
    count[1] = (double) nrules0;

    if (0 >= nrules0) {
      value = FALSE;
      goto done;
    }

    initialize_save_list(SUBWORD_RULE,1000,savell);

    navoid     = 1;
    avoid[0]   = subword0;

    initialize_dots('!',max(10,ntrials/50),50);

    for (ii = 0; ii < ntrials; ii++) {
      print_dots();
      random_overlap(subword0,subword0,maxlen,avoid,navoid,
                     count);
      if (abort() ) break;
    }

    end_dots();
    cout << "New rules found:  " << nsave << endl;

    if (nsave > 0) {
      if (use_heuristics) heuristics();
      (void) reduce_all_subsystems(SUBWORD_RULE);
      append_save_list(subword0);
      if (subword0->origin_in_use() ) subword0->make_origin();
      subword0->make_index();

      if (NULL != prefix0) {
        nrules0 = prefix0->get_nrules();
        if (0 >= nrules0) goto rtnfalse;
        if (NULL == subword0) goto rtnfalse;

        est_num = max(100,nrules0/2);
        initialize_save_list(PREFIX_RULE,est_num,savell);
        (void) reduce_all_subsystems(PREFIX_RULE);
        append_save_list(prefix0);
        prefix0->make_index();
      }

      value = TRUE;
      goto done;
    }

    else {
      append_save_list(subword0);
      value = FALSE;
    }
  }

  // We must consider prefix rules.

  else if (PREFIX_RULE == rtype) {
    if (NULL == prefix0) {
      cout << "There are no prefix rules." << endl;
      value = FALSE;
      goto done;
    }

    nrules0 = prefix0->get_nrules();
    count[1] = (double) nrules0;

    if (0 >= nrules0 || NULL == subword0) {
      value = FALSE;
      goto done;
    }

    est_num = max(100,nrules0/2);
    initialize_save_list(PREFIX_RULE,est_num,savell);

    navoid     = 1;
    avoid[0]   = subword0;

    initialize_dots('!',max(10,ntrials/50),50);

    for (ii = 0; ii < ntrials; ii++) {
      print_dots();
      random_overlap(prefix0,subword0,maxlen,avoid,navoid,
                     count);
      if (abort() ) break;
    }

    end_dots();

    if (nsave > 0) {
      (void) reduce_all_subsystems(PREFIX_RULE);
      append_save_list(prefix0);
      prefix0->make_index();
      value = TRUE;
      goto done;
    }

    else {
      append_save_list(prefix0);  // This just clears save list.
      value = FALSE;
      goto done;
    }
  }

 done:
  if (1 <= print_level) {
    cout << "Nodes in overlap tree:\n" << endl;

    for (j = 1; j <= maxlen; j++) {
      num = count[j];
      if (1 < j) num /= ntrials;
      cout << setw(4) << j << "  " << setw(12) << num;

      if (0.0 == num) {
        cout << endl;
        break;
      }

      else if (2 == j%3) cout << endl;
      else cout << "   ";
    }
  }

  fxfree( (void*) count);
  return value;

 rtnfalse:
  fxfree( (void*) count);
  return FALSE;
}



/////////////////////////////////////////////////////////////////
//
//              rewriting_system::compare_packed_rules
//
/////////////////////////////////////////////////////////////////

long rewriting_system::compare_packed_rules(
  sort_type    stp,
  packed_rule  v_rule,
  packed_rule  w_rule
)
{
  long    lex;

  switch(stp) {
    case LEN:
      if (v_rule.left.length == w_rule.left.length)
        return 0;

      else if (v_rule.left.length < w_rule.left.length)
        return -1;

      else return 1;

    case LEX:
      return compare_lex(v_rule.left, w_rule.left);

    case LENLENLEXLEX:
      if (v_rule.left.length != w_rule.left.length)
        if (v_rule.left.length < w_rule.left.length)
          return -1;

        else return -1;

      if (v_rule.right.length != w_rule.right.length)
        if (v_rule.right.length < w_rule.right.length)
          return -1;

        else return 1;

      if (0 != (lex = compare_lex(v_rule.left,w_rule.left) ) )
        return lex;

      return compare_lex(v_rule.right,w_rule.right);


    case LENLEXLENLEX:
      if (v_rule.left.length != w_rule.left.length)
        if (v_rule.left.length < w_rule.left.length)
          return -1;

        else return -1;

      if (0 != (lex = compare_lex(v_rule.left,w_rule.left) ) )
        return lex;

      if (v_rule.right.length != w_rule.right.length)
        if (v_rule.right.length < w_rule.right.length)
          return -1;

        else return 1;

      return compare_lex(v_rule.right,w_rule.right);
  }

  return 0;  // This is unnecessary, but it makes the compiler
             // happy.
}



/////////////////////////////////////////////////////////////////
//
//                  rewriting_system::compare_lex
//
/////////////////////////////////////////////////////////////////

long rewriting_system::compare_lex(
  word_reference  u_ref,
  word_reference  v_ref
)
{
  long           i,llu,llv,llm,gu,gv;
  gen_extractor  u_ext(u_ref.subsyst);
  gen_extractor  v_ext(v_ref.subsyst);

  llu = u_ref.length;
  llv = v_ref.length;
  llm = min(llu,llv);

  u_ext.reset(u_ref.origin,u_ref.bit_pos);
  v_ext.reset(v_ref.origin,v_ref.bit_pos);

  for (i = 0; i < llm; i++) {
    gu = u_ext.next();
    gv = v_ext.next();

    if (gu != gv)
      if (gu < gv) return -1;
      else return 1;
  }

  if (llu != llv)
    if (llu < llv) return -1;
    else return 1;

  return 0;
}



/////////////////////////////////////////////////////////////////
//
//                    rewriting_system::write
//
/////////////////////////////////////////////////////////////////

void rewriting_system::write(
  char  *file_name
)
{
  long       i,j,wt0,wt1,lv0,lv1,ninv,num,nsub;
  char       file[50];
  subsystem *subsyst;
  fstream    stream;
  
  if (strlen(file_name) > 0) {
    if (strlen(file_name) > 45)
      error("Name too long in rewriting_system::write");
    strcpy(file,file_name);
    strcat(file,".sys");
    stream.open(file, ios::out);
  }
  
  else system_file.open(stream);
  
  if (patterns) stream << "xsystem" << endl;
  else stream << "system" << endl;

  stream << "word_type:  ";
  
  if (a_type == MONOID) stream << "monoid" << endl;
  else stream << "group" << endl;

  if (a_type == GROUP) {  
    stream << "ngens:  " << n_gens << endl;

    if (patterns)
      stream << "nxgens:  " << n_xgens-n_gens << endl;

    stream << "naming_convention:  none" << endl;
    stream << "numbering_convention:  none" << endl;
    stream << "inverse_convention:  none" << endl;
    stream << "generators:" << endl;
    
    for (i = 0; i < t_xgens; i++) {
      gens[i].print_name(stream);
      stream << endl;
    }
    
    stream << "inverse_pairs:" << endl;
    
    for (i = 0; i < t_xgens; i++) {
      j = gens[i].get_inv_numb();
      
      if (i < j) {
        gens[i].print_name(stream);
        stream << "  ";
        gens[j].print_name(stream);
        stream << endl;
      }
    }
  } // a_type == GROUP
  
  else { // a_type == MONOID
    stream << "ngens:  " << n_gens << endl;
    if (patterns) stream << "nxgens:  " << n_xgens-n_gens << endl;
    stream << "naming_convention:  none" << endl;
    stream << "generators:" << endl;
    
    for (i = 0; i < n_gens; i++) {
      gens[i].print_name(stream);
      stream << endl;
    }
    
    stream << "ninv:  ";
    
    ninv = 0;
    
    for (i = 0; i < n_xgens; i++) 
      if (gens[i].get_invertible() )  ninv++;
      
    stream << ninv << endl;

    if (ninv > 0) {
      stream << "inverse_pairs:" << endl;

      for (i = 0; i < n_xgens; i++)
        if (gens[i].get_invertible() ) {
          gens[i].print_name(stream);
          stream << "  ";
          print_inverse(stream,gens[i]);
          stream << endl;
        }
    }
  }  // a_type == MONOID
  
  stream << "weight_convention:  ";
  
  if (n_gens == 0) stream << "none" << endl;
  
  else {
    wt1 = wt0 = gens[0].get_weight();
    
    for (i = 1; i < t_xgens; i++)
      if (wt0 != (wt1 = gens[i].get_weight() ) ) break;
      
    if (wt1 == wt0) stream << "constant" << endl;
    
    else {
      stream << "none" << endl;
      stream << "weights:" << endl;
      
      for (i = 0; i < t_xgens; i++)
        stream << gens[i].get_weight() << endl;
    }
  }
  
  stream << "level_convention:  ";
  
  if (n_gens == 0) stream << "none" << endl << "levels:" << endl;
  
  else {
    lv1 = lv0 = gens[0].get_level();
    
    for (i = 1; i < t_xgens; i++)
      if (lv0 != (lv1 = gens[i].get_level() ) )break;
      
    if (lv1 == lv0) stream << "constant" << endl;
    
    else {
      if (WRLNLX == o_type || WRWTLNLX == o_type)
        stream << "none" << endl;
      else stream << "rnone" << endl;

      stream << "levels:" << endl;
      
      for (i = 0; i < t_xgens; i++)
        stream << gens[i].get_level() << endl;
    }
  }

  nsub = n_sub;
  if (broken_prefix_rules != NULL) nsub++;
  if (broken_subword_rules != NULL) nsub++;

  stream << "nsub:  " << nsub << endl;
  stream << "subsystems:  " << endl;
  
  subsyst = first_subsystem;
  num     = 1;
  
  while (subsyst != NULL) {
    subsyst->write(stream,file_name,num++);
    subsyst = subsyst->get_next_subsystem();
  }
  
  if (broken_prefix_rules != NULL)
    broken_prefix_rules->write(stream,file_name,num++);

  if (broken_subword_rules != NULL)
    broken_subword_rules->write(stream,file_name,num++);

  stream.close();
}



/////////////////////////////////////////////////////////////////
//
//                 rewriting_system::print_word
//
/////////////////////////////////////////////////////////////////

void rewriting_system::print_word(
  ostream       &stream,
  external_form  ext_form,
  vector_long   &word
)
{
  long   i,j,len,wbits,zbits,nchar,nb;
  ulong  z,w;
  
  len = word.get_length();
  
  if (ext_form == LEN_NAMES   ||
      ext_form == LEN_NUMBERS ||
      ext_form == LEN_6BIT      )
    stream << len << " ";
    
  if (ext_form == NAMES    ||
      ext_form == LEN_NAMES  )
    if (len == 0) stream << "@";
    
    else
      for (i = 0; i < len; i++)
        gens[word[i]].print_name(stream);
  
  else if (ext_form == LEN_NUMBERS)
    for (i = 0; i < len; i++) {
      stream << word[i];
      if (i < len-1) stream << " ";
    }
    
  else if (ext_form == LEN_6BIT)
    if (len == 0) stream << "@";
    
    else {
      nchar = (n_bits*len+5)/6;
      j = 0;
      z = w = 0;
      zbits = wbits = 0;
      
      for (i = 0; i < nchar; i++) {
        while (zbits < 6) {
          if (wbits == 0)
            if (j >= len) {
              z = z<<(6-zbits);
              zbits = 6;
            }
            
            else {
              w     = word[j++];
              wbits = n_bits;
            }
            
          if (wbits+zbits <= 6) {
            z      = (z<<wbits)|(w&mask[wbits]);
            zbits += wbits;
            wbits  = 0;
          }
          
          else {
            nb     = 6 - zbits;
            z      = (z<<nb)|((w>>(wbits-nb))&mask[nb]);
            zbits  = 6;
            wbits -= nb;
          }
        }
        
        stream << (char) (z+48);
        zbits = z = 0;
      }
    }
}



/////////////////////////////////////////////////////////////////
//
//                 rewriting_system::print_counts
//
/////////////////////////////////////////////////////////////////

void rewriting_system::print_counts()
{
  cout << endl;
  
  cout << "n_rewrite      = " << setw(12) << n_rewrite <<
       "   total_input    = " << setw(12) << total_input << endl;
  
  cout << "n_applications = " << setw(12) << n_applications <<
       "   total_left     = " << setw(12) << total_left << endl;
       
  cout << "noverlaps      = " << setw(12) << noverlaps <<
       "   nbalance       = " << setw(12) << nbalance << endl;

  cout << "n_rabin_karp   = " << setw(12) << n_rabin_karp <<
       "   n_rk_tries     = " << setw(12) << n_rk_tries << endl;

  cout << "n_rk_hits      = " << setw(12) << n_rk_hits << endl;
}



/////////////////////////////////////////////////////////////////
//
//                       ask_rewrite_verbose
//
/////////////////////////////////////////////////////////////////

void rewriting_system::ask_rewrite_verbose(void)
{
  cout << "Do you want rewrite reporting turned on (y/n):  ";

  if (YES) set_rewrite_verbose(TRUE);
  else set_rewrite_verbose(FALSE);

  return;
}



/////////////////////////////////////////////////////////////////
//
//             rewriting_system::save_broken_rule
//
/////////////////////////////////////////////////////////////////

void rewriting_system::save_broken_rule(
  rule  &rul
)
{
  if (!use_broken) return;

  if (rul.r_type == SUBWORD_RULE) {
    if (broken_subword_rules == NULL)
      broken_subword_rules =
            new subsystem(this,SUBWORD_RULE,NO_INDEX);

    broken_subword_rules->add_rule(rul,FALSE,FALSE);
  }

  else if (rul.r_type == PREFIX_RULE) {
    if (broken_prefix_rules == NULL)
      broken_prefix_rules =
            new subsystem(this,PREFIX_RULE,NO_INDEX);

    broken_prefix_rules->add_rule(rul,FALSE,FALSE);
  }

  else if (PREFIX_PATTERN  == rul.r_type ||
           SUBWORD_PATTERN == rul.r_type   ) {
    // Identities should not occur as broken rules.
  }

  return;
}



/////////////////////////////////////////////////////////////////
//
//             rewriting_system::save_inverses
//
/////////////////////////////////////////////////////////////////

void rewriting_system::save_inverses(void)
{
  long      i,j,linv;
  rule      rul;

  rul.left.set_max_length(20);
  rul.right.set_length(0);

  for (i = 0; i < t_gens; i++)
    if(gens[i].get_invertible() ) {
      linv = gens[i].get_inv_length();
      rul.left.set_max_length(linv+1);
      rul.left.set_length(linv+1);
      rul.left.set_entry(0,i);

      for (j = 0; j < linv; j++)
         rul.left.set_entry(j+1,gens[i].get_inv_term(j) );

      save_rule(rul);
    }

  return;
}



/////////////////////////////////////////////////////////////////
//
//                 rewriting_system::details
//
/////////////////////////////////////////////////////////////////

void rewriting_system::details(
  long   i
)
{
  long        j,first,last;
  subsystem  *subsyst;

  if (i <= 0) {
    first = 1;
    last  = n_sub;
  }

  else first = last = i;

  subsyst = first_subsystem;

  for (j = 1; j < first; j++)
    if (subsyst != NULL) subsyst = subsyst->get_next_subsystem();

  for (j = first; j <= last; j++) {
    if (subsyst == NULL) {
      cout << "Null subsystem." << endl;
      return;
    }

    subsyst->details();
    cout << endl;
    subsyst = subsyst->get_next_subsystem();
  }

  return;
}



/////////////////////////////////////////////////////////////////
//
//                 rewriting_system::probe
//
/////////////////////////////////////////////////////////////////

// If ntrials is negative, printing of results is suppressed.

long rewriting_system::probe(
  bool            subword_only,
  long            ntrials,
  long            depth
)
{
  long         i,j,k,ng;
  double      *count,num,sum;
  bool         report;
  vector_long  wrd,genlist;
  packed_rule  rul;
  rule_reference  exclude = {NULL,0};
  subsystem   *subsyst;

  if (ntrials < 0) {
    report = FALSE;
    ntrials = -ntrials;
  }

  else report = TRUE;

  wrd.set_max_length(depth);
  genlist.set_max_length(t_gens);

  count = (double *) fxmllc( (depth+1)*sizeof(double) );

  if (count == NULL) {
    cout << "Unable to allocate space for counts" << endl;
    return 0;
  }

  count[0] = ntrials;
  for (j = 1; j <= depth; j++) count[j] = 0.0;

  subsyst = first_subsystem;

  while (subsyst != NULL) {
    subsyst->check_aux(REWRITE,subword_only,depth);
    subsyst = subsyst->get_next_subsystem();
  }

  for (i = 0; i < ntrials; i++) {
    wrd.set_length(0);
    num = 1.0;

    for (j = 0; j < depth; j++) {
      ng = 0;

      for (k = 0; k < t_gens; k++) {
        wrd.set_entry(j,k);
        wrd.set_length(j+1);

        subsyst = first_subsystem;

        while (subsyst != NULL) {
          subsyst->incremental_test(REWRITE,subword_only,
                            wrd,exclude,rul);

          if (rul.left.subsyst != NULL) break;

          subsyst = subsyst->get_next_subsystem();
        }

        if (subsyst != NULL) continue;  // wrd is reducible.

        genlist.set_entry(ng++,k);
      }

      if (0 == ng) break;
      wrd.set_entry(j,genlist[irand(0)%ng]);

      subsyst = first_subsystem;

      while (subsyst != NULL) {
        subsyst->incremental_test(REWRITE,subword_only,wrd,
                                  exclude,rul);
        subsyst = subsyst->get_next_subsystem();
      }

      num *= ng;
      count[j+1] += num;
    }
  }

  sum = 0.0;

  for (j = 0; j <= depth; j++) {
    num = count[j]/ntrials;
    sum += num;

    if (num < 0.5) break;

    if (report) {
      cout << setw(4) << j << "  " << setw(12) << num;

      if (2 == j%3) cout << endl;
      else cout << "   ";
    }
  }

  if (report)
    cout << endl << endl << "The total is " << sum << endl;

  fxfree( (void *)count );

  if (sum >100000000.0) return 100000000;

  return (long) sum;
}



/////////////////////////////////////////////////////////////////
//
//                rewriting_system::rewrite_rules
//
/////////////////////////////////////////////////////////////////

void rewriting_system::rewrite_rules(void)
{
  long              code,numequal,numbig,numsaved,numnotred;
  long              numout,nbits;
  char              str[500];
  bool              changed,reduced,echo;
  external_form     ext;
  rule_type         r_type,save_type;
  rule_reference    exclude = {NULL,0};
  rule              rul;
  subsystem        *subsyst,*subsyst0;
  fstream           fstrm_in,fstrm_out;

  nsave_total = 0;
  echo = get_echo();

  numequal = numnotred = numbig = numsaved = numout = 0;
  rul.set_max_length(bsize);

 get_in_file:
  cout <<
    "\nEnter name of input file (or \"stop\"):  ";

  cin >> str;
  if (echo) cout << str << endl;
  if ( !cin.ignore(10000,'\n') ) goto input_error;

  if( 0 == strcmp(str,"stop") ) return;

  if (!file_exists(str) ) {
    cout << "File does not exist" << endl;
    if (echo) exit(1);
    goto get_in_file;
  }

  fstrm_in.open(str, ios::in);

 get_type:
  cout <<
    "Enter rule type (subword or prefix):  ";

  cin >> str;
  if (echo) cout << str << endl;
  if ( !cin.ignore(10000,'\n') ) goto input_error;

  if      (0 == strcmp(str,"subword") ) r_type = SUBWORD_RULE;
  else if (0 == strcmp(str,"prefix") )  r_type = PREFIX_RULE;

  else {
    cout << "Invalid type" << endl;
    if (echo) exit(1);
    goto get_type;
  }

  rul.r_type = r_type;

 get_format:
  nbits = n_bits;

  cout << "Enter external format" << endl <<
          "(names, len_names, len_numbers, or len_6bit):  ";

  cin >> str;
  if (echo) cout << str << endl;
  if ( !cin.ignore(10000,'\n') ) goto input_error;

  if (0 == strcmp(str,"names") )
    ext = NAMES;

  else if (0 == strcmp(str,"len_names") )
    ext = LEN_NAMES;

  else if (0 == strcmp(str,"len_numbers") )
    ext = LEN_NUMBERS;

  else if (0 == strcmp(str,"len_6bit") ) {
    ext = LEN_6BIT;
    cout << "\nEnter value of nbits:  ";
    scan_long(cin,nbits);
  }

  else {
    cout << "Invalid format" << endl;
    if (echo) exit(1);
    goto get_format;
  }

 get_out_file:
  cout <<
    "\nEnter name of output file (or \"none\" or \"stop\"):  ";

  cin >> str;
  if (echo) cout << str << endl;
  if ( !cin.ignore(10000,'\n') ) goto input_error;

  if (0 == strcmp(str,"stop") ) return;

  subsyst = NULL;

  if (0 == strcmp(str,"none") ) {
    cout << "\nEnter max saved left side (positive integer):  ";
    cin  >> savell;
    if (echo) cout << savell << endl;
    if ( !cin.ignore(10000,'\n') ) goto input_error;

    cout << "\nEnter max saved right side (nonnegative integer):  ";
    cin  >> savelr;
    if (echo) cout << savelr << endl;
    if (!cin.ignore(10000,'\n') ) goto input_error;

   get_save_type:
    cout <<
    "\nEnter type for saving (same or subword_pattern):  ";

    cin >> str;
    if (get_echo() ) cout << str << endl;
    if ( !cin.ignore(10000,'\n') ) goto input_error;

    if      (0 == strcmp(str,"same") )
      save_type = r_type;

    else if (0 == strcmp(str,"subword_pattern") ) {
      if (SUBWORD_RULE != r_type) {
        cout <<
        "Only subword rules can be promoted to subword patterns."
             << endl;

        if (echo) exit(1);
        goto get_save_type;
      }

      if (t_xgens < 2*t_gens) {
        cout << "There are not enough pattern generators." << endl;
        if (echo) exit(1);
        goto get_save_type;
      }

      save_type = SUBWORD_PATTERN;
    }

    else {
      cout << "Invalid type" << endl;
      if (echo) exit(1);
      goto get_save_type;
    }

    subsyst = first_subsystem;

    while (subsyst != NULL && save_type != subsyst->get_r_type() )
      subsyst = subsyst->get_next_subsystem();

    if (NULL == subsyst) {
      if (save_type == SUBWORD_PATTERN)
        subsyst = new subsystem(this,save_type,NO_INDEX);
      else
        subsyst = new subsystem(this,save_type,HYBRID);

      add_subsystem(subsyst);
    }

    subsyst0 = subsyst;
  }

  else if (file_exists(str) ) {
    cout << "File exists already" << endl;
    if (echo) exit(1);
    goto get_out_file;
  }

  if (NULL != subsyst) {
    // Argument 1000 in next line is a guess!

    if (save_type != SUBWORD_PATTERN)
      initialize_save_list(save_type,1000,savell);
  }

  else fstrm_out.open(str, ios::out);

  if (!echo) initialize_loop();

  for (;;) {
    if (!echo) count_loop();
    if (abort() ) break;
    code = read_word(fstrm_in,ext,nbits,rul.left);
    if (1 == code) break;
    if (2 == code) goto input_error;

    code = read_word(fstrm_in,ext,nbits,rul.right);

    if (1 == code || 2 == code) goto input_error;

    rewrite(rul,exclude,changed,reduced);

    if (orient_rule(rul) ) {
      numequal++;
      continue;
    }

    if (!reduced) numnotred++;
    else balance_rule(rul);

    if (NULL == subsyst) {
      numout++;
      print_word(fstrm_out,ext,rul.left);
      fstrm_out << " ";
      print_word(fstrm_out,ext,rul.right);
      fstrm_out << endl;
    }

    else if (reduced)
      if (check_rule(rul) ) {
        if (SUBWORD_PATTERN == save_type) {
          promote(rul);
          subsyst->add_rule(rul,FALSE,FALSE);
        }

        else save_rule(rul);

        numsaved++;
      }

      else numbig++;
  }

  if (!echo) end_loop();

  fstrm_in.close();

  if (NULL == subsyst) {
    fstrm_out.close();
    cout << "Number of equalities:  " << numequal << endl;
    cout << "Number not reduced:  " << numnotred << endl;
    cout << "Number written out:  " << numout << endl;
  }

  else {
    cout << "Number of equalities:  " << numequal << endl;
    cout << "Number not reduced:  " << numnotred << endl;
    cout << "Number too big:  " << numbig << endl;
    cout << "Number saved:  " << numsaved << endl;

    if (SUBWORD_PATTERN != save_type) {
      if (numsaved > 0) {
        if (use_heuristics) heuristics();
        reduce_all_subsystems(save_type);
        append_save_list(subsyst0);
        if (subsyst0->origin_in_use() ) subsyst0->make_origin();
        subsyst0->make_index();
      }

      else append_save_list(subsyst0);  // Clears save list!
    }
  }

  return;

 input_error:
  cout << "Input error" << endl;
  return;
}



/////////////////////////////////////////////////////////////////
//
//                rewriting_system::add_powers
//
/////////////////////////////////////////////////////////////////

void rewriting_system::add_powers(
  long        exponent,
  long        maxlen,
  long        firstg,
  long        lastg
)
{
  long                 exp2,lw,nsub;
  long                 g,i,j,k,npow,nfail;
  bool                 changed,reduced;
  vector_long          word;
  rule                 rul;
  packed_rule          p_rule;
  rule_reference       exclude = {NULL,0};
  subsystem           *subsyst,*subsyst1;

  if (0 > firstg || firstg > lastg || lastg >= t_gens) {
    cout << "Error in generator numbers" << endl;
    if (get_echo() ) exit(1);
    return;
  }

  nsave_total = 0;
  nfail = 0;

  if (print_level >= 1) {
    cout << "Adding powers, exponent = " << exponent
         << ", maxlen = " << maxlen << endl;

    if (logll > 0)
      log_stream << "Adding powers, exponent = " << exponent
                 << ", maxlen = " << maxlen << endl << flush;
  }

  update_time();

  rul.set_max_length(max(bsize,exponent*maxlen) );

  // Locate the first subword subsystem.

  subsyst1 = first_subsystem;

  while (subsyst1 != NULL &&
         SUBWORD_RULE != subsyst1->get_r_type() )
    subsyst1 = subsyst1->get_next_subsystem();

  if (subsyst1 == NULL) {
    cout << "No subword system in add_powers." << endl;
    return;
  }

  exp2 = (exponent*maxlen)/2;
  initialize_save_list(SUBWORD_RULE,1000,savell);
  nsub = n_sub;

  // Check that subsystems have the auxiliary space they need.

  subsyst = first_subsystem;

  while (subsyst != NULL) {
    subsyst->check_aux(BACKTRACK,TRUE,maxlen);
    subsyst = subsyst->get_next_subsystem();
  }

  // Begin backtrack search for reduced words of length at most
  // maxlen.

  npow = probe(TRUE,-200,maxlen);

  if (print_level >= 2)
    cout << "Estimated number of powers:  " << npow << endl;

  initialize_dots('^',max(10,npow/50),50);

  word.set_max_length(maxlen);
  lw = 1;
  word.set_length(lw);
  word.set_entry(0,firstg);

  while (0 < word.get_length() ) {
    if (abort() ) break;

    subsyst = first_subsystem;

    while (subsyst != NULL) {
      subsyst->incremental_test(BACKTRACK,TRUE,word,exclude,
                                p_rule);
      if (p_rule.left.subsyst != NULL) goto increment;
      subsyst = subsyst->get_next_subsystem();
    }

    print_dots();
    rul.right.set_length(0);
    rul.left.set_length(lw*exponent);
    k = 0;

    for (i = 0; i < exponent; i++)
      for (j = 0; j < lw; j++)
        rul.left.set_entry(k++,word[j]);

    rewrite(rul,exclude,changed,reduced);

    if (!reduced) nfail++;

    else if (!orient_rule(rul) ) {
      balance_rule(rul);

      if (check_rule(rul) ) save_rule(rul);
      else nfail++;

      if (nsub < n_sub) {
        nsub++;
        subsyst = last_subsystem;
        subsyst->check_aux(BACKTRACK,TRUE,maxlen);

        for (i = 1; i <= lw; i++) {
          word.set_length(i);
          subsyst->incremental_test(BACKTRACK,TRUE,word,exclude,
                                    p_rule);
          if (p_rule.left.subsyst != NULL) break;
        }

        if (i <= lw) {
          lw = i;
          goto increment;
        }
      }
    }

    if (lw >= maxlen) goto increment;

    word.set_entry(lw++,firstg);
    word.set_length(lw);
    continue;

   increment:
    while(lw > 0 && (g = word[lw-1]) == lastg) lw--;
    word.set_length(lw);
    if (lw <= 0) continue;
    word.set_entry(lw-1,g+1);
  }  // while (0 < word.get_length() )

  end_dots();

  // If rules were saved, reduce all subword rules and append
  // save list to subsyst1.

  cout << "The number of new rules is " << nsave << endl;
  cout << "The number of failures is " << nfail << endl;

  if (nsave > 0) {
    if (use_heuristics) heuristics();
    (void) reduce_all_subsystems(SUBWORD_RULE);
    append_save_list(subsyst1);
    subsyst1->make_index();
  }

  else append_save_list(subsyst1); // This just clears save list

  return;
}


/////////////////////////////////////////////////////////////////
//
//                rewriting_system::add_engel
//
/////////////////////////////////////////////////////////////////

void rewriting_system::add_engel(
  long        engelnum,
  long        firstg,
  long        lastg,
  long        maxlen
)
{
  long                 exp2,lw,le,x,y;
  long                 g,i,j,k,nfail,nengel,lw1,lw2;
  bool                 changed,reduced;
  vector_long          word1,word2,engelword;
  rule                 rul;
  rule_reference       exclude = {NULL,0};
  subsystem           *subsyst,*subsyst1;

  if (0 >= engelnum || 0 >= maxlen) return;

  nsave_total = 0;
  nfail = 0;

  if (print_level >= 1) {
    cout << "Adding Engel rules, Engel number = " << engelnum
         << ", maxlen = " << maxlen << endl;

    if (logll > 0)
      log_stream << "Adding Engel rules, Engel number = " << engelnum
             << ", maxlen = " << maxlen << endl << flush;
  }

  update_time();

  engelword.set_max_length(1);
  engelword.set_length(1);
  engelword.set_entry(0,2);

  for (i = 0; i < engelnum; i++) {
    le = engelword.get_length();
    engelword.set_max_length(2*(le+1) );
    engelword.set_length(2*(le+1) );
    engelword.set_entry(2*le+1,1);

    for (j = 0; j < le; j++)
      engelword.set_entry(j+le+1,engelword[j]);

    engelword.set_entry(le,-1);

    for (j = 0; j < le/2; j++) {
      x = engelword[j];
      y = engelword[le-1-j];
      engelword.set_entry(j,-y);
      engelword.set_entry(le-1-j,-x);
    }

    if (1 == le) engelword.set_entry(0,-engelword[0]);

    engelword.free_reduce();
  }

  engelword.cyclic_reduce();
  le = engelword.get_length();

  // Locate the first subword subsystem.

  subsyst1 = first_subsystem;

  while (subsyst1 != NULL &&
         SUBWORD_RULE != subsyst1->get_r_type() )
    subsyst1 = subsyst1->get_next_subsystem();

  if (subsyst1 == NULL) {
    cout << "No subword system in add_engel." << endl;
    return;
  }

  initialize_save_list(SUBWORD_RULE,1000,savell);

  nengel = 1;
  for (i = 0; i < maxlen; i++) nengel *= (lastg + 1 - firstg);
  nengel = nengel*nengel;
  initialize_dots('^',max(10,nengel/50),50);

  word1.set_max_length(maxlen);
  word2.set_max_length(maxlen);
  word2.set_length(0);

  while (word2.next(maxlen,firstg,lastg) ) {
    if (abort() ) break;
    lw2 = word2.get_length();
    word1.set_length(0);

    while (word1.next(maxlen,firstg,lastg) ) {
      if (abort() ) break;
      lw1 = word1.get_length();
      print_dots();
      rul.right.set_length(0);
      rul.left.set_length(0);

      for (i = 0; i < le; i++) {
        x = engelword[i];

        if      (1  == x) rul.left.append(word1);
        else if (-1 == x) append_inverse(rul.left,word1);
        else if (2  == x) rul.left.append(word2);
        else              append_inverse(rul.left,word2);
      }

      rewrite(rul,exclude,changed,reduced);

      if (!reduced) nfail++;

      else if (!orient_rule(rul) ) {
        balance_rule(rul);

        if (check_rule(rul) ) save_rule(rul);
        else nfail++;

      }
    }
  }

  end_dots();

  // If rules were saved, reduce all subword rules and append
  // save list to subsyst1.

  cout << "The number of new rules is " << nsave << endl;
  cout << "The number of failures is " << nfail << endl;

  if (nsave > 0) {
    if (use_heuristics) heuristics();
    (void) reduce_all_subsystems(SUBWORD_RULE);
    append_save_list(subsyst1);
    subsyst1->make_index();
  }

  else append_save_list(subsyst1); // This just clears save list

  return;
}


/////////////////////////////////////////////////////////////////
//
//              rewriting_system::add_random_powers
//
/////////////////////////////////////////////////////////////////

void rewriting_system::add_random_powers(
  long        ntrials,
  long        exponent,
  long        maxlen,
  long        firstg,
  long        lastg
)
{
  long                 lw,nsub;
  long                 g,i,j,k,npow,ii,jj,ng,nfail;
  bool                 changed,reduced;
  vector_long          word,genlist;
  rule                 rul;
  packed_rule          p_rule;
  rule_reference       exclude = {NULL,0};
  subsystem           *subsyst,*subsyst1;

  nsave_total = 0;
  nfail = 0;

  if (print_level >= 1) {
    cout << "Adding random powers, exponent = " << exponent
         << ", maxlen = " << maxlen
         << ", ntrials = " << ntrials << endl;

    if (logll > 0)
      log_stream << "Adding random powers, exponent = " << exponent
                 << ", maxlen = " << maxlen
                 << ", ntrials = " << ntrials << endl << flush;
  }

  update_time();

  genlist.set_max_length(t_gens);
  rul.set_max_length(max(bsize,exponent*maxlen) );

  // Locate the first subword subsystem.

  subsyst1 = first_subsystem;

  while (subsyst1 != NULL &&
         SUBWORD_RULE != subsyst1->get_r_type() )
    subsyst1 = subsyst1->get_next_subsystem();

  if (subsyst1 == NULL) {
    cout << "No subword system in add_random_powers." << endl;
    return;
  }

  initialize_save_list(SUBWORD_RULE,1000,savell);
  nsub = n_sub;

  // Check that subsystems have the auxiliary space they need.

  subsyst = first_subsystem;

  while (subsyst != NULL) {
    subsyst->check_aux(BACKTRACK,TRUE,maxlen);
    subsyst = subsyst->get_next_subsystem();
  }

  // Begin backtrack search for reduced words of length at most
  // maxlen.

  word.set_max_length(maxlen);
  initialize_dots('@',max(10,ntrials/50),50);

  for (ii = 0; ii < ntrials; ii++) {
    if (abort() ) break;
    print_dots();
    word.set_length(0);

    for (j = 0; j < maxlen; j++) {
      ng = 0;

      for (k = firstg; k <= lastg; k++) {
        word.set_entry(j,k);
        word.set_length(j+1);

        subsyst = first_subsystem;

        while (subsyst != NULL) {
          subsyst->incremental_test(BACKTRACK,TRUE,
                            word,exclude,p_rule);

          if (p_rule.left.subsyst != NULL) break;

          subsyst = subsyst->get_next_subsystem();
        }

        if (subsyst != NULL) continue;  // word is reducible.

        genlist.set_entry(ng++,k);
      }

      if (0 == ng) break;
      word.set_entry(j,genlist[irand(0)%ng]);

      subsyst = first_subsystem;

      while (subsyst != NULL) {
        subsyst->incremental_test(BACKTRACK,TRUE,word,
                                  exclude,p_rule);
        subsyst = subsyst->get_next_subsystem();
      }

      rul.right.set_length(0);
      rul.left.set_length( (j+1)*exponent);
      k = 0;

      for (i = 0; i < exponent; i++)
        for (jj = 0; jj <= j; jj++)
          rul.left.set_entry(k++,word[jj]);

      rewrite(rul,exclude,changed,reduced);

      if (!reduced) nfail++;

      else if (!orient_rule(rul) ) {
        balance_rule(rul);

        if (check_rule(rul) ) save_rule(rul);
        else nfail++;

        if (nsub < n_sub) {
          nsub++;
          subsyst = last_subsystem;
          subsyst->check_aux(BACKTRACK,TRUE,maxlen);

          for (i = 1; i <= j+1; i++) {
            word.set_length(i);
            subsyst->incremental_test(BACKTRACK,TRUE,word,exclude,
                                      p_rule);
            if (p_rule.left.subsyst != NULL) break;
          }

          if (i <= j+1) break;
        }
      }   // if (!orient_rule
    }     // for (j = 0;
  }       // for (ii = 0;

  end_dots();

  // If rules were saved, reduce all subword rules and append
  // save list to subsyst1.

  cout << "The number of new rules is " << nsave << endl;
  cout << "The number of failures is " << nfail << endl;

  if (nsave > 0) {
    if (use_heuristics) heuristics();
    (void) reduce_all_subsystems(SUBWORD_RULE);
    append_save_list(subsyst1);
    subsyst1->make_index();
  }

  else append_save_list(subsyst1); // This just clears save list

  return;
}



/////////////////////////////////////////////////////////////////
//
//              rewriting_system::add_subword_powers
//
/////////////////////////////////////////////////////////////////

void rewriting_system::add_subword_powers(
  long        exponent
)
{
  long                 lw,code;
  long                 g,i,j,k,npow,ii,jj,ng,nfail;
  bool                 changed,reduced;
  vector_long          word;
  rule                 rul;
  rule_reference       exclude = {NULL,0};
  subsystem           *subsyst1;

  cout << "Enter the word in names format:" << endl;

  code = read_word(cin,NAMES,0,word);

  if (1 == code || 2 == code) {
    cout << "Error reading word." << endl;
    return;
  }

  lw = word.get_length();

  nsave_total = 0;
  nfail = 0;

  if (print_level >= 1) {
    cout << "Adding subword powers, exponent = " << exponent << endl;

    if (logll > 0)
      log_stream << "Adding subword powers, exponent = " << exponent
                 << endl << flush;
  }

  // Locate the first subword subsystem.

  subsyst1 = first_subsystem;

  while (subsyst1 != NULL &&
         SUBWORD_RULE != subsyst1->get_r_type() )
    subsyst1 = subsyst1->get_next_subsystem();

  if (subsyst1 == NULL) {
    cout << "No subword system in add_subword_powers." << endl;
    return;
  }

  update_time();

  rul.set_max_length(max(bsize,exponent*lw) );

  initialize_save_list(SUBWORD_RULE,1000,savell);

  for (j = 1; j <= lw; j++)
    for (i = 0; i < j; i++) {
      rul.right.set_length(0);
      rul.left.set_length((j-i)*exponent);

      ii = 0;
      for (jj = 0; jj < exponent; jj++)
        for (k = i; k < j; k++)
          rul.left.set_entry(ii++,word[k]);

      rewrite(rul,exclude,changed,reduced);

      if (!reduced) nfail++;

      else if (!orient_rule(rul) ) {
        balance_rule(rul);

        if (check_rule(rul) ) save_rule(rul);
        else nfail++;
      }   // if (!orient_rule
    }     // for (i = 0;
          // for (j = 0;

  // If rules were saved, reduce all subword rules and append
  // save list to subsyst1.

  cout << "The number of new rules is " << nsave << endl;
  cout << "The number of failures is " << nfail << endl;

  if (nsave > 0) {
    if (use_heuristics) heuristics();
    (void) reduce_all_subsystems(SUBWORD_RULE);
    append_save_list(subsyst1);
    subsyst1->make_index();
  }

  else append_save_list(subsyst1); // This just clears save list

  return;
}



/////////////////////////////////////////////////////////////////
//
//             rewriting_system::rewrite_words -- void
//
/////////////////////////////////////////////////////////////////

void rewriting_system::rewrite_words(void)
{
  char           file[100],str[100];
  long           nbits;
  external_form  ext;
  bool           exists,subword_only;
  ifstream       instream;
  ofstream       outstream;

  exists = FALSE;

  while (!exists) {
    cout <<
    "\nEnter input file (or \"none\" or \"stop\"):  ";
    scan_string(cin,file);

    if (0 == strcmp(file,"stop") ) return;
    if (0 == strcmp(file,"none") ) break;

    exists = file_exists(file);
    if (!exists) cout << "file does not exist." << endl;
  }

  for (;;) {
    cout << "Enter external format" << endl <<
          "(names, len_names, len_numbers, or len_6bit):  ";

    cin >> str;
    if (get_echo() ) cout << str << endl;
    if ( !cin.ignore(10000,'\n') ) error("Input error");

    if (0 == strcmp(str,"names") )
      ext = NAMES;

    else if (0 == strcmp(str,"len_names") )
      ext = LEN_NAMES;

    else if (0 == strcmp(str,"len_numbers") )
      ext = LEN_NUMBERS;

    else if (0 == strcmp(str,"len_6bit") ) {
      ext = LEN_6BIT;
      cout << "Enter value of nbits:  ";
      scan_long(cin,nbits);
    }

    else {
      cout << "Invalid format" << endl;
      continue;
    }

    break;
  }

  for (;;) {
    cout << "Enter type of rewriting (subword or prefix):  ";
    cin  >> str;
    if (get_echo() ) cout << str << endl;
    if ( !cin.ignore(10000,'\n') ) error("Input error");

    if (0 == strcmp(str,"subword") )
      subword_only = TRUE;

    else if (0 == strcmp(str,"prefix") )
      subword_only = FALSE;

    else {
      cout << "Invalid type" << endl;
      continue;
    }

    break;
  }

  if (0 != strcmp(file,"none") )
    for (;;) {
      cout <<
        "Enter name of output file (or \"stop\"):  ";

      cin >> str;
      if (get_echo() ) cout << str << endl;

      if ( !cin.ignore(10000,'\n') ) {
        instream.close();
        return;
      }

      if (0 == strcmp(str,"stop") ) {
        instream.close();
        return;
      }

      if (file_exists(str) ) {
        cout << "File exists" << endl;
        continue;
      }

      outstream.open(str, ios::out);
      break;
    }

  if (0 == strcmp(file,"none") ) {
    cout << "Signal end with empty word" << endl;
    rewrite_words(cin,ext,nbits,subword_only,TRUE,cout);
  }

  else {
    instream.open(file, ios::in);
    rewrite_words(instream,ext,nbits,subword_only,FALSE,outstream);
    instream.close();
    outstream.close();
  }

  return;
}



/////////////////////////////////////////////////////////////////
//
//             rewriting_system::rewrite_words -- stream
//
/////////////////////////////////////////////////////////////////

void rewriting_system::rewrite_words(
  istream       &instream,
  external_form  ext,
  long           nbits,
  bool           subword_only,
  bool           stop_on_empty,
  ostream       &outstream
)
{
  long           code;
  vector_long    word;
  bool           changed,reduced,echo;
  rule_reference exclude = {NULL,0};

  word.set_max_length(1000);

  echo = get_echo();

  if (!echo) initialize_loop();

  for (;;) {
    if (!echo) count_loop();
    code = read_word(instream,ext,nbits,word);
    if (code == 1 || code == 2) break;

    if (stop_on_empty && 0 == word.get_length() ) break;

    rewrite_from_left(
          subword_only,word,word,exclude,changed,reduced);

    if (!reduced) outstream << "Result not reduced." << endl;
    print_word(outstream,ext,word);
    outstream << endl;
  }

  if (!echo) end_loop();
  return;
}



/////////////////////////////////////////////////////////////////
//
//                rewriting_system::log_on
//
/////////////////////////////////////////////////////////////////

void rewriting_system::log_on(
  long           maxlen,
  char          *file,
  external_form  form
)
{
  if (maxlen <= 0) {
    cout << "maxlen must be positive for logging." << endl;
    return;
  }

  if (logll > 0) {
    // logging is already on.

    cout <<
      "Logging is currently on." << endl;
    cout << "Only maxlen and format have been changed."
      << endl;

    logll = maxlen;
    logform = form;

    return;
  }

  logll = maxlen;
  logform = form;

  if (file_exists(file) ) log_stream.open(file, ios::app);
  else log_stream.open(file, ios::out);
}



/////////////////////////////////////////////////////////////////
//
//               rewriting_system::log_off
//
/////////////////////////////////////////////////////////////////

void rewriting_system::log_off(void)
{
  if (logll <= 0) {
    cout << "Logging is not currently on." << endl;
    return;
  }

  logll = 0;
  log_stream.close();
}



/////////////////////////////////////////////////////////////////
//
//                    rewriting_system::promote
//
/////////////////////////////////////////////////////////////////

void rewriting_system::promote(
  rule  &rul
)
{
  long    i;

  for (i = 0; i < rul.left.get_length(); i++)
    rul.left.set_entry(i,t_gens+rul.left[i]);

  for (i = 0; i < rul.right.get_length(); i++)
    rul.right.set_entry(i,t_gens+rul.right[i]);

  return;
}


/////////////////////////////////////////////////////////////////
//
//                rewriting_system::append_inverse
//
/////////////////////////////////////////////////////////////////

void rewriting_system::append_inverse(
  vector_long  &wrd,
  long          g
)
{
  if (!gens[g].invertible)
    error("Noninvertible generator in append_inverse");

  wrd.append(gens[g].inverse);
}



/////////////////////////////////////////////////////////////////
//
//                 rewriting_system::append_inverse
//
/////////////////////////////////////////////////////////////////

void rewriting_system::append_inverse(
  vector_long  &u,
  vector_long  &v
)
{
  int   lv,i,x;

  lv = v.get_length();

  for (i = lv-1; i >= 0; i--) {
    x = v[i];
    append_inverse(u,x);
  }
}



/////////////////////////////////////////////////////////////////
//
//                   rewriting_system::append_power
//
/////////////////////////////////////////////////////////////////

void rewriting_system::append_power(
  vector_long   &u,
  vector_long   &v,
  long           exp
)
{
  int  i;

  if (0 < exp)
    for (i = 0; i < exp; i++) u.append(v);
  else if (0 > exp)
    for (i = 0; i < -exp; i++) append_inverse(u,v);
}



/////////////////////////////////////////////////////////////////
//
//                   rewriting_system::decode
//
/////////////////////////////////////////////////////////////////

long rewriting_system::decode(
  vector_char &buff,
  long        &index,
  vector_long &word
)
{
  long                  sign,caret,n,g;
  char                  c;
  vector_long           u,v,com,last;

  sign  = 1;
  caret = 0;
  last.set_length(0);
  word.set_length(0);
  c   = buff[index];

  if ('\0' == c) goto done;

  if ('@' == c) {
    index++;
    goto done;
  }

  // A string to be decoded can't start with a caret, a comma,
  // or an integer.

  if ('-' == c || '^' == c || isdigit(c) || ',' == c)
    goto parse_error;

  for (;;) {
    c = buff[index];

    if (sign < 0) {
      get_integer(buff,index,n);
      if (0 == n) goto parse_error;
      n = -n;
      if ('^' == buff[index]) goto parse_error;
      append_power(word,last,n);
      last.set_length(0);
      caret = 0;
      sign  = 1;
    }

    else if ('^' == c) {
      if (caret) goto parse_error;

      caret = 1;
      if (0 == last.get_length() ) goto parse_error;
      index++;
    }

    else if (isdigit(c) ) {
      get_integer(buff,index,n);
      if (0 == n) goto parse_error;
      if ('^' == buff[index] || !caret) goto parse_error;
      if (0 == last.get_length() ) goto parse_error;
      append_power(word,last,n);
      last.set_length(0);
      caret = 0;
    }

    else if (isalpha(c) ) {
      if (0 != get_generator(buff,index,g) ) goto parse_error;

      if (caret) {
        if ('^' == buff[index]) goto parse_error;
        append_inverse(word,g);
        word.append(last);
        word.append(g);
        last.set_length(0);
        caret = 0;
      }

      else {
        if (0 < last.get_length() ) {
          word.append(last);
          last.set_length(0);
        }

        last.append(g);
        continue;
      }
    }

    else if ( ')' == c || ']' == c ||
             '\0' == c || ',' == c   ) goto done;

    else if ( '-' == c) {
      if (!caret) goto parse_error;
      sign = -1;
      index++;
    }

    else if ('(' == c) {
      index++;
      if (0 != decode(buff,index,u) ) goto parse_error;
      if (')' != buff[index]) goto parse_error;
      if (0 == u.get_length() ) goto parse_error;

      if (caret) {
        append_inverse(word,u);
        word.append(last);
        word.append(u);
        last.set_length(0);
        caret = 0;
      }

      else {
        word.append(last);
        last.copy(u);
      }

      index++;
    }

    else if ('[' == c) {
      index++;
      if (0 != decode(buff,index,com) ) goto parse_error;

      while (']' != buff[index]) {
        if (',' != buff[index]) goto parse_error;
        index++;
        if (0 != decode(buff,index,v) ) goto parse_error;
        u.copy(com);
        com.set_length(0);
        append_inverse(com,u);
        append_inverse(com,v);
        com.append(u);
        com.append(v);
      }

      word.append(last);
      last.copy(com);
      index++;
    }

    else goto parse_error;
  }

 done:
  if (caret) goto parse_error;
  if (0 < last.get_length() ) word.append(last);

  return 0;

 parse_error:
  cout << "Parse error in decode" << endl;
  cout << index << endl;
  word.print();
  cout << endl << buff << endl;
  return 2;
}


/////////////////////////////////////////////////////////////////
//
//                       get_integer
//
/////////////////////////////////////////////////////////////////

void get_integer(vector_char &buff, long  &index, long &n)
{
  char    c;

  n = 0;

  for (;;) {
    c = buff[index];

    if (!isdigit(c) ) return;
    n = 10*n + c - '0';
    index++;
  }
}


/////////////////////////////////////////////////////////////////
//
//                 rewriting_system::get_generator
//
/////////////////////////////////////////////////////////////////

long rewriting_system::get_generator(
  vector_char  &buff,
  long         &index,
  long         &g
)
{
  long    j,state,current,p,bit,h,old,index1;
  char    c;

  c = buff[index];

  if (!isalpha(c) ) {
    cout << "Names start with letters" << endl;
    return 2;
  }

  j       = 0;
  g       = LONG_MAX;
  state   = 0;
  current = 0;
  index1  = index;

  for (;;) {
    c = buff[index1];

    if ( '(' == c || ')'  == c || '[' == c || ']' == c ||
         ',' == c || '\0' == c || '^' == c) {
      if (g >= t_xgens) {
        cout << "Parse error in get_generator" << endl;
        return 2;
      }

      return 0;
    }

    if ( !isdigit(c) && !isalpha(c) ) {
      cout << "Invalid character in get_generator" << endl;
      return 2;
    }

    index1++;

    for (p = CHAR_BIT-2; p >= 0; p--) {
      bit   = (c>>p) & 1;
      state = autn.get_entry(state,bit);
      if (state == LONG_MAX) break;
    }

    if ( LONG_MAX != state &&
         LONG_MAX != (h = autn.get_entry(state,2) ) ) {
      //  The current prefix is a generator name.

      g   = h;
      index = index1;
    }

    if (state == LONG_MAX )
      if (g != LONG_MAX) {
        if (g >= t_xgens) {
          cout << "Invalid generator in get_generator" << endl;
          return 2;
        }

        return 0;
      }

      else {
        cout << "Parse error in get_generator" << endl;
        return 2;
      }
  }
}


/////////////////////////////////////////////////////////////////
//
//                 rewriting_system::add_instances
//
/////////////////////////////////////////////////////////////////

void rewriting_system::add_instances(
  long  subno,
  long  maxill,  // Maximum initial left side.
  long  maxilr,  // Maximum initial right side.
  long  maxlen,
  long  firstg,
  long  lastg
)
{
  long          nfail,nrules,orig,i,ll,lr;
  packed_rule   p_rule;
  rule          rul;
  subsystem    *subsyst,*subsyst1;

  if (use_patterns) {
    cout <<
      "Turn off pattern rewriting before constructing instances."
         << endl;
    if (get_echo() ) exit(1);
    return;
  }

  if (0 > firstg || firstg > lastg || lastg >= t_gens) {
    cout << "Error in generator numbers" << endl;
    if (get_echo() ) exit(1);
    return;
  }

  nsave_total = 0;
  subsyst = first_subsystem;

  while (subno > 1 && subsyst != NULL) {
    subsyst = subsyst->get_next_subsystem();
    subno--;
  }

  if (subno > 1 || subsyst == NULL) {
    cout << "Invalid subsystem number" << endl;
    if (get_echo() ) exit(1);
    return;
  }

  if (print_level >= 1) {
    cout << "Adding instances of patterns at " << (ulong)subsyst
         << ", maxlen = " << maxlen << endl;

    if (logll > 0)
      log_stream << "Add intances of patterns at " << (ulong)subsyst
                 << ", maxlen = " << maxlen << endl << flush;
  }

  update_time();

  rul.set_max_length(max(bsize,max(maxill,maxilr) ) );

  subsyst1 = first_subsystem;

  while (subsyst1 != NULL &&
         SUBWORD_RULE != subsyst1->get_r_type() )
    subsyst1 = subsyst1->get_next_subsystem();

  if (subsyst1 == NULL) {
    cout << "No subword system in add_instances." << endl;
    if (get_echo() ) exit(1);
    return;
  }

  initialize_save_list(SUBWORD_RULE,1000,savell);

  nfail = 0;
  nrules = subsyst->get_nrules();
  orig = 0;
  initialize_loop();

  for (i = 0; i < nrules; i++) {
    count_loop();
    if (abort() ) break;
    subsyst->rule_by_origin(orig,p_rule);
    ll = p_rule.left.length;
    lr = p_rule.right.length;
    if (ll > maxill || lr > maxilr) continue;
    subsyst->unpack(p_rule,rul);
    rule_instances(rul,maxill,maxilr, maxlen,firstg,lastg,
                   nfail);
  }

  end_loop();

  cout << "The number of new rules is " << nsave << endl;
  cout << "The number of failures is " << nfail << endl;

  if (nsave > 0) {
    if (use_heuristics) heuristics();
    (void) reduce_all_subsystems(SUBWORD_RULE);
    append_save_list(subsyst1);
    subsyst1->make_index();
  }

  else append_save_list(subsyst1); // This just clears save list

  return ;
}



/////////////////////////////////////////////////////////////////
//
//                rewriting_system::rule_instances
//
/////////////////////////////////////////////////////////////////

void rewriting_system::rule_instances(
  rule     &rul,
  long      maxill,
  long      maxilr,
  long      maxlen,
  long      firstg,
  long      lastg,
  long     &nfail
)
{
  long            ll,lr,vargen,varinv,lcount,rcount,i,x;
  long            maxlw,ll1,lr1,lw;
  bool            changed,reduced;
  vector_long     word;
  rule            rul1;
  rule_reference  exclude = {NULL,0};

  rul1.set_max_length(rul.right.get_max_length() );
  ll = rul.left.get_length();
  lr = rul.right.get_length();
  vargen = -1;
  varinv = -1;
  lcount = rcount = 0;

  for (i = 0; i < ll; i++) {
    x = rul.left[i];
    if (x < t_gens) continue;

    if (lcount == 0) {
      vargen = x;
      lcount = 1;

      if (a_type == GROUP) varinv = gens[x].get_inv_numb();
    }

    else if (x == vargen || x == varinv) lcount++;
  }

  if (lcount == 0) {
    count_loop();
    rewrite(rul,exclude,changed,reduced);

    if (!reduced) nfail++;

    else if (!orient_rule(rul) ) {
      balance_rule(rul);

      if (check_rule(rul) ) save_rule(rul);
      else nfail++;
    }
  }

  else {
    for (i = 0; i < lr; i++) {
      x = rul.right[i];
      if (x == vargen || x == varinv) rcount++;
    }

    maxlw = min(maxlen, 1 + (maxill - ll)/lcount);
    if (rcount > 0) maxlw = min(maxlw, 1+(maxilr-lr)/rcount);

    word.set_max_length(maxlw);

    while (word.next(maxlw,firstg,lastg) ) {
      if (abort() ) break;
      lw = word.get_length();
      rul1.left.set_length(0);
      rul1.right.set_length(0);

      for (i = 0; i < ll; i++) {
        x = rul.left[i];

        if (x != vargen && x != varinv) {
          ll1 = rul1.left.get_length();
          rul1.left.set_entry(ll1,x);
          rul1.left.set_length(ll1+1);
        }

        else if (x == vargen) rul1.left.append(word);
        else append_inverse(rul1.left,word);
      }

      for (i = 0; i < lr; i++) {
        x = rul.right[i];

        if (x != vargen && x != varinv) {
          lr1 = rul1.right.get_length();
          rul1.right.set_entry(lr1,x);
          rul1.right.set_length(lr1+1);
        }

        else if (x == vargen) rul1.right.append(word);
        else append_inverse(rul1.right,word);
      }

      rule_instances(rul1,maxill,maxilr,maxlen,firstg,lastg,nfail);
    }
  }
}



/////////////////////////////////////////////////////////////////
//
//                  rewriting_system::memory
//
/////////////////////////////////////////////////////////////////

void rewriting_system::memory(void)
{
  subsystem   *subsyst;

  subsyst = first_subsystem;

  while (subsyst != NULL) {
    subsyst->memory();
    subsyst = subsyst->get_next_subsystem();
  }
}



/////////////////////////////////////////////////////////////////
//
//                    rewriting_system::rk_update
//
/////////////////////////////////////////////////////////////////

void rewriting_system::rk_update(
  long    tries,
  long    hits
)
{
  subsystem  *subsyst;

  n_rabin_karp++;
  n_rk_tries  += tries;
  n_rk_hits   += hits;

  return;

/*  
  if (0 != (n_rabin_karp%10000) ) return;

  subsyst = first_subsystem;

  while (NULL != subsyst) {
    subsyst->hash_summary();
    subsyst = subsyst->get_next_subsystem();
  }
*/
}



/////////////////////////////////////////////////////////////////
//
//                rewriting_system::set_use_broken
//
/////////////////////////////////////////////////////////////////

void rewriting_system::set_use_broken(
  bool  val
)
{
  use_broken = val;

  if (val) return;

  if (NULL != broken_subword_rules) {
    delete broken_subword_rules;
    broken_subword_rules = NULL;
  }

  if (NULL != broken_prefix_rules) {
    delete broken_prefix_rules;
    broken_prefix_rules = NULL;
  }
}



/////////////////////////////////////////////////////////////////
//
//                   rewriting_system::cancel
//
/////////////////////////////////////////////////////////////////

void rewriting_system::cancel()
{
  long    g;

  cout << "Enter internal number of generator:  ";
  cin  >> g;

  if (get_echo() ) cout << g << endl;

  if ( !cin.ignore(10000,'\n') ) {
    cout << "Can't clear line" << endl;
    exit(1);
  }

  if (g < 0 || g >= t_gens)
    cout << "Invalid generator number" << endl;

  else
    gens[g].set_cancelable(TRUE);
}



/////////////////////////////////////////////////////////////////
//
//            rewriting_system::word_level
//
/////////////////////////////////////////////////////////////////

long rewriting_system::word_level(
   vector_long    &wrd
)
{
  long    lev,lw,i;

  lev = LONG_MAX;
  lw  = wrd.get_length();

  for (i = 0; i < lw; i++)
    lev = min(lev,gens[wrd[i]].get_level() );

  return lev;
}



/////////////////////////////////////////////////////////////////
//
//                    rewriting_system::write_rules
//
/////////////////////////////////////////////////////////////////

void rewriting_system::write_rules(
  long           numsub,
  char          *file_name,
  external_form  extfrm,
  ulong          llmin,
  ulong          llmax,
  ulong          lrmin,
  ulong          lrmax,
  ulong          genmin,
  ulong          genmax
)
{
  long       i,j,wt0,wt1,lv0,lv1,ninv,num,nsub;
  char       file[50];
  subsystem *subsyst;
  fstream    stream;
  
  if (file_name == NULL || 0 == strlen(file_name) ) {
    cout << "Invalid file name in rewriting_system::write_rules"
         << endl;
    return;
  }

  subsyst = first_subsystem;
  
  while (subsyst != NULL && numsub > 1) {
    subsyst = subsyst->get_next_subsystem();
    numsub--;
  }
  
  if (numsub > 1) {
    cout <<
    "Could not find subsystem in rewriting_system::write_rules"
         << endl;

    return;
  }

  stream.open(file_name, ios::out);

  subsyst->write_rules(stream,extfrm,llmin,llmax,lrmin,lrmax,
                        genmin,genmax);

  stream.close();
}



/////////////////////////////////////////////////////////////////
//
//           rewriting_system::reidemeister_schreier
//
/////////////////////////////////////////////////////////////////

void rewriting_system::reidemeister_schreier(void)
{
  long           first,last,g,numequal,numnotred,numbig,numout;
  long           cnumb,num,cur,lc,code,i,x,nbits = 0,offset;
  long           maxl,maxr,maxused;
  char           sep[100],name[100],ifile[100],ofile[100];
  bool           echo,changed,reduced;
  external_form  iext,oext;
  vector_long   *reps;
  rule_reference exclude = {NULL, 0};
  ifstream       istream;
  ofstream       ostream;
  rule           rul,rul2;

  rul.set_max_length(bsize);
  rul2.set_max_length(bsize);

  numequal = numnotred = numbig = numout = 0;

 get_sep:
  cout << "Enter separator generator:  ";
  cin  >> sep;

  if (get_echo() ) cout << sep << endl;
  if ( !cin.ignore(10000,'\n') ) goto clear_error;

  for (g = 0; g < n_gens; g++) {
    gens[g].get_name(name);
    if (0 == strcmp(sep,name) ) break;
  }

  if (g >= n_gens) {
    cout << "Invalid generator name" << endl;
    if (get_echo() ) exit(1);
    goto get_sep;
  }

  gens[g].set_cancelable(TRUE);

 get_first:
  cout << "Enter first generator (internal number):  ";
  cin >> first;
  if (get_echo() ) cout << first << endl;
  if ( !cin.ignore(10000,'\n') ) goto clear_error;

  if (first < 0 || first >= n_gens) {
    cout << "Invalid number" << endl;
    if (get_echo() ) exit(1);
    goto get_first;
  }

 get_last:
  cout << "Enter last generator (internal number):  ";
  cin >> last;

  if (get_echo() ) cout << last << endl;
  if ( !cin.ignore(10000,'\n') ) goto clear_error;

  if (last < first || last >= n_gens) {
    cout << "Invalid number" << endl;
    if (get_echo() ) exit(1);
    goto get_last;
  }

 get_cnumb:
  cout << "Enter maximum number of cosets:  ";
  cin  >> cnumb;
  if (get_echo() ) cout << cnumb << endl;
  if ( !cin.ignore(10000,'\n') ) goto clear_error;

  if (1 > cnumb) {
    cout << "Invalid number" << endl;
    if (get_echo() ) exit(1);
    goto get_cnumb;
  }

  reps = new vector_long[cnumb];
  reps[0].set_length(1);
  reps[0].set_entry(0,g);

  cur = 0;
  num = 1;

  while (cur < num) {
    lc = reps[cur].get_length();

    for (x = first; x <= last; x++) {
      if (num == cnumb) break;

      reps[num].set_max_length(lc+1);
      reps[num].set_copy(reps[cur]);
      reps[num].set_length(lc+1);
      reps[num].set_entry(lc,x);

      rewrite_from_left(
          TRUE,reps[num],reps[num],exclude,changed,reduced);

      if (!reduced || changed) continue;
      num++;
    }

    if (num == cnumb) break;
    cur++;
  }

  cout << num << " coset reps found." << endl;

 get_ifile:
  cout << "Enter input file name (or \"stop\"):  ";
  cin  >> ifile;
  if ( !cin.ignore(10000,'\n') ) goto clear_error;

  if (get_echo() ) cout << ifile << endl;

  if (0 == strcmp(ifile,"stop") ) goto done;

  if (!file_exists(ifile) ) {
    cout << "File does not exist" << endl;
    if (get_echo() ) exit(1);
    goto get_ifile;
  }

 get_iform:
  cout << "Enter external format" << endl <<
        "(names, len_names, len_numbers, or len_6bit):  ";

  cin >> name;
  if ( !cin.ignore(10000,'\n') ) goto clear_error;

  if (get_echo() ) cout << name << endl;

  offset = 0;

  if (0 == strcmp(name,"names") )
    iext = NAMES;

  else if (0 == strcmp(name,"len_names") )
    iext = LEN_NAMES;

  else if (0 == strcmp(name,"len_numbers") )
    iext = LEN_NUMBERS;

  else if (0 == strcmp(name,"len_6bit") ) {
    iext = LEN_6BIT;
    cout << "Enter value of nbits:  ";
    scan_long(cin,nbits);
    cout << "Enter offset:  ";
    scan_long(cin,offset);
  }

  else {
    cout << "Invalid format" << endl;
    goto get_iform;
  }

 get_maxused:
  cout << "Enter maximum length for use:  ";
  cin  >> maxused;
  if ( !cin.ignore(1000,'\n') ) goto clear_error;
  if (get_echo() ) cout << maxused << endl;
  if (maxused <= 0) goto done;

 get_ofile:
  cout << "Enter output file name (or \"stop\"):  ";
  cin  >> ofile;
  if ( !cin.ignore(10000,'\n') ) goto clear_error;

  if (get_echo() ) cout << ofile << endl;

  if (0 == strcmp(ofile,"stop") ) goto done;

  if (file_exists(ofile) ) {
    cout << "File exists" << endl;
    if (get_echo() ) exit(1);
    goto get_ofile;
  }

 get_oform:
  cout << "Enter external format" << endl <<
        "(names, len_names, len_numbers, or len_6bit):  ";

  cin >> name;
  if ( !cin.ignore(10000,'\n') ) goto clear_error;

  if (get_echo() ) cout << name << endl;

  if (0 == strcmp(name,"names") )
    oext = NAMES;

  else if (0 == strcmp(name,"len_names") )
    oext = LEN_NAMES;

  else if (0 == strcmp(name,"len_numbers") )
    oext = LEN_NUMBERS;

  else if (0 == strcmp(name,"len_6bit") )
    oext = LEN_6BIT;

  else {
    cout << "Invalid format" << endl;
    goto get_oform;
  }

 get_maxl:
  cout << "Enter maximum saved left side:  ";
  cin  >> maxl;

  if ( !cin.ignore(10000,'\n') ) goto clear_error;

  if (get_echo() ) cout << maxl << endl;

  if (maxl <= 0) {
    cout << "Invalid number" << endl;
    goto get_maxl;
  }

 get_maxr:
  cout << "Enter maximum saved right side:  ";
  cin  >> maxr;

  if ( !cin.ignore(10000,'\n') ) goto clear_error;

  if (get_echo() ) cout << maxr << endl;

  if (maxr < 0) {
    cout << "Invalid number" << endl;
    goto get_maxr;
  }

  istream.open(ifile,ios::in);
  ostream.open(ofile,ios::out);

  echo = get_echo();
  if (!echo) initialize_loop();

  for (;;) {
    if (!echo) count_loop();
    code = read_word(istream,iext,nbits,rul.left);
    if (1 == code) break;
    if (2 == code) goto input_error;

    code = read_word(istream,iext,nbits,rul.right);

    if (1 == code || 2 == code) goto input_error;

    if (maxused < rul.left.get_length() ||
        maxused < rul.right.get_length()  ) continue;

    rul.left.add_offset(offset);
    rul.right.add_offset(offset);

    for (i = 0; i < num; i++) {
      rul2.left.copy(reps[i]);
      rul2.right.copy(reps[i]);
      rul2.left.append(rul.left);
      rul2.right.append(rul.right);
      rewrite(rul2,exclude,changed,reduced);

      if (orient_rule(rul2) ) {
        numequal++;
        continue;
      }

      if (!reduced) numnotred++;

      else if (rul2.left.get_length() > maxl ||
               rul2.right.get_length() > maxr  )
        numbig++;

      else {
        balance_rule(rul2);
        numout++;
        print_word(ostream,oext,rul2.left);
        ostream << " ";
        print_word(ostream,oext,rul2.right);
        ostream << endl;
      }
    }
  }

  if (!echo) end_loop();

  istream.close();
  ostream.close();

  cout << "Number of equalities:  " << numequal << endl;
  cout << "Number not reduced:  " << numnotred << endl;
  cout << "Number too big:  " << numbig << endl;
  cout << "Number written out:  " << numout << endl;

 done:
  delete[] reps;

  return;

 input_error:
  cout << "Input error reading file in reidemeister_schreier"
       << endl;

  exit(1);

 clear_error:
  cout << "Unable to clear command line\n";
  exit(1);
}



/////////////////////////////////////////////////////////////////
//
//                 rewriting_system::conjugate
//
/////////////////////////////////////////////////////////////////

void rewriting_system::conjugate(
  long     num_iter
)
{
  long             llmax,lrmax,ll2,lr2,nrules,orig,i,x,iter;
  bool             changed,reduced;
  packed_rule      p_rule;
  rule             rule1,rule2;
  rule_reference   exclude = {NULL,0};
  subsystem       *subsyst;

  nsave_total = 0;
  subsyst = first_subsystem;

  while (subsyst != NULL) {
    if (PREFIX_RULE == subsyst->get_r_type() ) break;
    subsyst = subsyst->get_next_subsystem();
  }

  if (subsyst == NULL || 0 == (nrules = subsyst->get_nrules() ) ) {
    cout << "The are no prefix rules" << endl;
    return;
  }

  iter = 0;

  while (!abort() && num_iter != 0) {
    nrules = subsyst->get_nrules();
    iter++;
    if (print_level >= 2)
      cout << "Pass " << iter << ".  There are "<< nrules
           << " rules." << endl;

    llmax = subsyst->get_llmax();
    lrmax = subsyst->get_lrmax();

    if (bsize < 1 + llmax || bsize < 1 + lrmax) {
      cout << "The buffer size is too small" << endl;
      return;    
    }

    rule1.set_max_length(bsize);
    rule2.set_max_length(bsize);
    rule2.set_r_type(PREFIX_RULE);

    initialize_save_list(PREFIX_RULE,nrules,savell);
    orig = 0;

    for (i = 0; i < nrules; i++) {
      if (abort() ) break;
      subsyst->rule_by_origin(orig,p_rule);
      subsyst->unpack(p_rule,rule1);

      for (x = 0; x < t_gens; x++) {
        rule2.left.set_length(1);
        rule2.left.set_entry(0,x);
        rule2.right.set_length(1);
        rule2.right.set_entry(0,x);
        rule2.left.append(rule1.left);
        rule2.right.append(rule1.right);

        rewrite(rule2,exclude,changed,reduced);
        if ( orient_rule(rule2) || !reduced ) continue;
        balance_rule(rule2);

        ll2 = rule2.left.get_length();
        lr2 = rule2.right.get_length();

        if (ll2 <= savell && lr2 <= savelr) save_rule(rule2);
      }
    }

    if (nsave > 0) {
      if (use_heuristics) heuristics();
      (void) reduce_all_subsystems(PREFIX_RULE);
      append_save_list(subsyst);
      if (subsyst->origin_in_use() ) subsyst->make_origin();
      subsyst->make_index();
      if (num_iter > 0) num_iter--;
    }

    else {
      append_save_list(subsyst);
      num_iter = 0;
    }
  }

  return;
}



/////////////////////////////////////////////////////////////////
//
//                 rewriting_system::comp
//
/////////////////////////////////////////////////////////////////

long rewriting_system::comp(
  vector_long &u,
  vector_long &v
)
{
  if (compare == LNLX) return lnlx(u,v);
  else if (compare == WTLNLX) return wtlnlx(u,v);
  else if (compare == WRLNLX) return wrlnlx(u,v);
  else if (compare == WRWTLNLX) return wrwtlnlx(u,v);
  else if (compare == RWRLNLX)  return rwrlnlx(u,v);
  else if (compare == RWRWTLNLX) return rwrwtlnlx(u,v);
}


/////////////////////////////////////////////////////////////////
//
//                 rewriting_system::count
//
/////////////////////////////////////////////////////////////////

void rewriting_system::count(
  vector_long    &wrd,
  long            maxlen
)
{
  long            i,j,k,ng,g,lw,status,origlen;
  double         *cnt,num,sum;
  bool            print,subword_only,echo;
  char            str[100];
  rule_type       rtp;
  packed_rule     p_rule;
  rule_reference  exclude = {NULL,0};
  external_form   ext;
  subsystem      *subsyst;
  fstream         fstrm_out;

  update_time();
  echo = get_echo();

  cout << "Do you want to count subword irreducible words:  ";

  if (YES) subword_only = TRUE;
  else subword_only = FALSE;

  cout << "\nDo you want the words written to a file:  ";

  if (YES) {
    print = TRUE;

   get_file:
    cout << "\nEnter name of file (or \"stop\"):  ";
    cin >> str;
    if (echo) cout << str << endl;

    if ( !cin.ignore(10000,'\n') ) goto input_error;
    if ( 0 == strcmp(str,"stop") ) return;

    if (file_exists(str) ) {
      cout << "File does not exist" << endl;
      if (echo) exit(1);
      goto get_file;
    }

    fstrm_out.open(str, ios::out);

   get_format:
    cout << "Enter external format" << endl <<
            "(names, len_names, len_numbers, or len_6bit):  ";

    cin >> str;
    if (echo) cout << str << endl;
    if ( !cin.ignore(10000,'\n') ) goto input_error;

    if (0 == strcmp(str,"names") )
      ext = NAMES;

    else if (0 == strcmp(str,"len_names") )
      ext = LEN_NAMES;

    else if (0 == strcmp(str,"len_numbers") )
      ext = LEN_NUMBERS;

    else if (0 == strcmp(str,"len_6bit") )
      ext = LEN_6BIT;

    else {
      cout << "Invalid format" << endl;
      if (echo) exit(1);
      goto get_format;
    }
  }

  else print = FALSE;

  origlen = wrd.get_length();
  wrd.set_max_length(maxlen);

  cnt = (double *) fxmllc( (maxlen+1)*sizeof(double) );

  if (cnt == NULL) {
    cout << "Unable to allocate space for counts" << endl;
    return;
  }

  for (j = 0; j <= maxlen; j++) cnt[j] = 0.0;

  subsyst = first_subsystem;

  while (subsyst != NULL) {
    subsyst->check_aux(BACKTRACK,subword_only,maxlen);
    subsyst = subsyst->get_next_subsystem();
  }

  for (lw = 1; lw <= origlen; lw++) {
    wrd.set_length(lw);

    subsyst = first_subsystem;

    while (NULL != subsyst) {
      subsyst->incremental_test(BACKTRACK,subword_only,wrd,
                                exclude,p_rule);
      if (p_rule.left.subsyst != NULL) {
        cout << "The initial word is reducible." << endl;
        goto done;
      }

      subsyst = subsyst->get_next_subsystem();
    }
  }

  cnt[origlen] = 1.0;

  if (maxlen <= origlen) goto done;

  // Begin backtrack search for nonempty irreducible words.

  lw = origlen + 1;
  wrd.set_length(lw);
  wrd.set_entry(lw-1,0);

  for (;;) {
    subsyst = first_subsystem;

    while (NULL != subsyst) {
      subsyst->incremental_test(BACKTRACK,subword_only,wrd,
                                exclude,p_rule);

      if (p_rule.left.subsyst != NULL) goto increment;
      subsyst = subsyst->get_next_subsystem();
    }

    cnt[lw] += 1.0;

    if (print) {
      print_word(fstrm_out,ext,wrd);
      fstrm_out << endl;
    }

    if (lw == maxlen) goto increment;

    wrd.set_entry(lw++,0);
    wrd.set_length(lw);
    continue;

   increment:
    while (lw > origlen && (g = wrd[lw-1]) == t_gens-1) lw--;
    if (lw <= origlen) break;
    wrd.set_entry(lw-1,g+1);
    wrd.set_length(lw);
  }
    
 done:
  sum = 0.0;
  cout << "The numbers of irreducible words are:" << endl << endl;

  for (i = origlen; i <= maxlen; i++) {
    cout << setw(4) << i << "  " << setw(12) << setprecision(0)
         << setiosflags(ios::fixed) << cnt[i] << endl;
    sum += cnt[i];

    if (cnt[i] < 1.0) break;
  }
  
  cout << endl << endl;
  cout << "The total number of words is " << sum << endl;

  fxfree( (void *)cnt );

  return;

 input_error:
  cout << "Error clearing line" << endl;
  exit(1);
}



/////////////////////////////////////////////////////////////////
//
//                 rewriting_system::countx
//
/////////////////////////////////////////////////////////////////

void rewriting_system::countx(
  vector_long    &wrd,
  long            maxlen
)
{
  long            i,j,k,ng,g,lw,status,origlen,lw1;
  double         *cnt,num,sum,*back,*same,*forward;
  bool            print,subword_only,echo,changed,reduced;
  char            str[100];
  rule_type       rtp;
  vector_long     wrd1;
  packed_rule     p_rule;
  rule_reference  exclude = {NULL,0};
  external_form   ext;
  subsystem      *subsyst;
  fstream         fstrm_out;

  update_time();
  echo = get_echo();

  if (this->o_type != LNLX) {
    cout << "This command only makes sense when words are ordered\n";
    cout << "by the lenlex ordering." << endl;
    return;
  }

  cout << "Do you want to count subword irreducible words:  ";

  if (YES) subword_only = TRUE;
  else subword_only = FALSE;

  cout << "\nDo you want the words written to a file:  ";

  if (YES) {
    print = TRUE;

   get_file:
    cout << "\nEnter name of file (or \"stop\"):  ";
    cin >> str;
    if (echo) cout << str << endl;

    if ( !cin.ignore(10000,'\n') ) goto input_error;
    if ( 0 == strcmp(str,"stop") ) return;

    if (file_exists(str) ) {
      cout << "File does not exist" << endl;
      if (echo) exit(1);
      goto get_file;
    }

    fstrm_out.open(str, ios::out);

   get_format:
    cout << "Enter external format" << endl <<
            "(names, len_names, len_numbers, or len_6bit):  ";

    cin >> str;
    if (echo) cout << str << endl;
    if ( !cin.ignore(10000,'\n') ) goto input_error;

    if (0 == strcmp(str,"names") )
      ext = NAMES;

    else if (0 == strcmp(str,"len_names") )
      ext = LEN_NAMES;

    else if (0 == strcmp(str,"len_numbers") )
      ext = LEN_NUMBERS;

    else if (0 == strcmp(str,"len_6bit") )
      ext = LEN_6BIT;

    else {
      cout << "Invalid format" << endl;
      if (echo) exit(1);
      goto get_format;
    }
  }

  else print = FALSE;

  origlen = wrd.get_length();
  wrd.set_max_length(maxlen);

  cnt = (double *) fxmllc( (maxlen+1)*sizeof(double) );
  back = (double *) fxmllc( (maxlen+1)*sizeof(double) );
  same = (double *) fxmllc( (maxlen+1)*sizeof(double) );
  forward = (double *) fxmllc( (maxlen+1)*sizeof(double) );

  if (cnt == NULL || back == NULL || same == NULL || forward == NULL) {
    cout << "Unable to allocate space for counts" << endl;
    return;
  }

  for (j = 0; j <= maxlen; j++) {
    cnt[j] = 0.0;
    back[j] = 0.0;
    same[j] = 0.0;
    forward[j] = 0.0;
  }

  rewrite_from_left(subword_only,wrd,wrd,exclude,changed,reduced);

  if (changed || !reduced) {
    cout << "The initial word is reducible." << endl;
    goto done;
  }

  cnt[origlen] = 1.0;

  if (maxlen <= origlen) goto done;

  // Begin backtrack search for nonempty irreducible words.

  wrd1.set_max_length(maxlen);

  lw = origlen + 1;
  wrd.set_length(lw);
  wrd.set_entry(lw-1,0);

  for (;;) {
    rewrite_from_left(subword_only,wrd,wrd1,exclude,changed,reduced);
    lw1 = wrd1.get_length();

    if (reduced && !changed) {
      cnt[lw] += 1.0;
      forward[lw-1] += 1.0;

      if (print) {
        print_word(fstrm_out,ext,wrd);
        fstrm_out << endl;
      }

      if (lw == maxlen) goto increment;

      wrd.set_entry(lw++,0);
      wrd.set_length(lw);
      continue;
    }

    else {
      if (lw1 == lw) forward[lw-1] += 1.0;
      else if (lw1 == lw-1) same[lw-1] += 1.0;
      else if (lw1+2 == lw) back[lw-1] += 1.0;
    }

   increment:
    while (lw > origlen && (g = wrd[lw-1]) == t_gens-1) lw--;
    if (lw <= origlen) break;
    wrd.set_entry(lw-1,g+1);
    wrd.set_length(lw);
  }
    
 done:
  sum = 0.0;
  cout << "The numbers of irreducible words are:" << endl << endl;

  for (i = origlen; i <= maxlen; i++) {
    cout << setw(4) << i << "  " << setw(12) << setprecision(0)
         << setiosflags(ios::fixed) << cnt[i]
         << setw(12) << back[i] << setw(12) << same[i]
         << setw(12) << forward[i] << endl;
    sum += cnt[i];

    if (cnt[i] < 1.0) break;
  }
  
  cout << endl << endl;
  cout << "The total number of words is " << sum << endl;

  fxfree( (void *)cnt );
  fxfree( (void *)back );
  fxfree( (void *)same );
  fxfree( (void *)forward);

  return;

 input_error:
  cout << "Error clearing line" << endl;
  exit(1);
}




/////////////////////////////////////////////////////////////////
//
//                 rewriting_system::countg
//
/////////////////////////////////////////////////////////////////

void rewriting_system::countg(
  vector_long    &wrd,
  long            maxlen
)
{
  long            i,j,k,ng,g,lw,status,origlen,lw1;
  double         *cnt,num,sum,*back,*same,*forward;
  bool            print,subword_only,echo,changed,reduced;
  char            str[100];
  rule_type       rtp;
  vector_long     wrd1;
  packed_rule     p_rule;
  rule_reference  exclude = {NULL,0};
  external_form   ext;
  subsystem      *subsyst;
  fstream         fstrm_out;

  update_time();
  echo = get_echo();

  if (this->o_type != LNLX) {
    cout << "This command only makes sense when words are ordered\n";
    cout << "by the lenlex ordering." << endl;
    return;
  }

  subword_only = TRUE;

  cout << "\nDo you want the words written to a file:  ";

  if (YES) {
    print = TRUE;

   get_file:
    cout << "\nEnter name of file (or \"stop\"):  ";
    cin >> str;
    if (echo) cout << str << endl;

    if ( !cin.ignore(10000,'\n') ) goto input_error;
    if ( 0 == strcmp(str,"stop") ) return;

    if (file_exists(str) ) {
      cout << "File does not exist" << endl;
      if (echo) exit(1);
      goto get_file;
    }

    fstrm_out.open(str, ios::out);

   get_format:
    cout << "Enter external format" << endl <<
            "(names, len_names, len_numbers, or len_6bit):  ";

    cin >> str;
    if (echo) cout << str << endl;
    if ( !cin.ignore(10000,'\n') ) goto input_error;

    if (0 == strcmp(str,"names") )
      ext = NAMES;

    else if (0 == strcmp(str,"len_names") )
      ext = LEN_NAMES;

    else if (0 == strcmp(str,"len_numbers") )
      ext = LEN_NUMBERS;

    else if (0 == strcmp(str,"len_6bit") )
      ext = LEN_6BIT;

    else {
      cout << "Invalid format" << endl;
      if (echo) exit(1);
      goto get_format;
    }
  }

  else print = FALSE;

  origlen = wrd.get_length();
  wrd.set_max_length(maxlen);

  cnt = (double *) fxmllc( (maxlen+1)*sizeof(double) );
  back = (double *) fxmllc( (maxlen+1)*sizeof(double) );
  same = (double *) fxmllc( (maxlen+1)*sizeof(double) );
  forward = (double *) fxmllc( (maxlen+1)*sizeof(double) );

  if (cnt == NULL || back == NULL || same == NULL || forward == NULL) {
    cout << "Unable to allocate space for counts" << endl;
    return;
  }

  for (j = 0; j <= maxlen; j++) {
    cnt[j] = 0.0;
    back[j] = 0.0;
    same[j] = 0.0;
    forward[j] = 0.0;
  }

  rewrite_from_left(subword_only,wrd,wrd,exclude,changed,reduced);

  if (changed || !reduced) {
    cout << "The initial word is reducible." << endl;
    goto done;
  }

  cnt[origlen] = 1.0;

  if (maxlen <= origlen) goto done;

  // Begin backtrack search for nonempty irreducible words.

  wrd1.set_max_length(maxlen);

  lw = origlen + 1;
  wrd.set_length(lw);
  wrd.set_entry(lw-1,0);

  for (;;) {
    rewrite_from_left(subword_only,wrd,wrd1,exclude,changed,reduced);
    lw1 = wrd1.get_length();

    if (reduced && lw1 == lw) {
      cnt[lw] += 1.0;
      forward[lw-1] += 1.0;

      if (print) {
        print_word(fstrm_out,ext,wrd);
        fstrm_out << endl;
      }

      if (lw == maxlen) goto increment;

      wrd.set_entry(lw++,0);
      wrd.set_length(lw);
      continue;
    }

    else {
      if (lw1 == lw) forward[lw-1] += 1.0;
      else if (lw1 == lw-1) same[lw-1] += 1.0;
      else if (lw1+2 == lw) back[lw-1] += 1.0;
    }

   increment:
    while (lw > origlen && (g = wrd[lw-1]) == t_gens-1) lw--;
    if (lw <= origlen) break;
    wrd.set_entry(lw-1,g+1);
    wrd.set_length(lw);
  }
    
 done:
  sum = 0.0;
  cout << "The numbers of geodesics are:" << endl << endl;

  for (i = origlen; i <= maxlen; i++) {
    cout << setw(4) << i << "  " << setw(12) << setprecision(0)
         << setiosflags(ios::fixed) << cnt[i]
         << setw(12) << back[i] << setw(12) << same[i]
         << setw(12) << forward[i] << endl;
    sum += cnt[i];

    if (cnt[i] < 1.0) break;
  }
  
  cout << endl << endl;
  cout << "The total number of geodesics is " << sum << endl;

  fxfree( (void *)cnt );
  fxfree( (void *)back );
  fxfree( (void *)same );
  fxfree( (void *)forward);

  return;

 input_error:
  cout << "Error clearing line" << endl;
  exit(1);
}




/////////////////////////////////////////////////////////////////
//
//                 rewriting_system::add_subs
//
/////////////////////////////////////////////////////////////////

void rewriting_system::add_subs(
  long  maxill,  // Maximum initial left side.
  long  maxilr,  // Maximum initial right side.
  long  maxtll,  // Maximum left side after substitution
  long  maxtlr,  // Maximum right side after substitution
  long  maxlen,
  long  firstg,
  long  lastg
)
{
  long          nfail,nrules,orig,i,ll,lr;
  bool          invcode;
  packed_rule   p_rule;
  rule          rul;
  subsystem    *subsyst,*subsyst1;

  if (0 > firstg || firstg > lastg || lastg >= t_gens) {
    cout << "Error in generator numbers" << endl;
    if (get_echo() ) exit(1);
    return;
  }

  if (print_level >= 1) {
    cout << "Adding instances of subword rules"
         << ", maxlen = " << maxlen << endl;

    if (logll > 0)
      log_stream << "Add intances of subword rules"
                 << ", maxlen = " << maxlen << endl << flush;
  }

  nsave_total = 0;
  update_time();

  if (t_gens != t_xgens && t_xgens < 2*t_gens) {
    cout << "The number of extra generators is not adequate."
         << endl;

    if (get_echo() ) exit(1);
    return;
  }

  if (t_gens != t_xgens)
    for (i = 0; i < t_gens; i++) {
      invcode = gens[i].get_invertible();

      if (invcode != gens[i+t_gens].get_invertible() ) {
        cout << "Inconsistency in inverses" << endl;
        if (get_echo() ) exit(1);
        return;
      }

      if (invcode && a_type == GROUP)
        if (gens[i].get_inv_numb()+t_gens !=
             gens[i+t_gens].get_inv_numb()    ) {
          cout << "Inconsistency in inverse numbering" << endl;
          if (get_echo() ) exit(1);
          return;
        }
    }

  else
    for (i = 0; i < t_gens; i++) {
      invcode = gens[i].get_invertible();
      gens[i+t_gens].set_invertible(invcode);

      if (invcode && a_type == GROUP)
        gens[i+t_gens].set_inv_numb(t_gens+gens[i].get_inv_numb() );
    }

  rul.set_max_length(max(bsize,max(maxtll,maxtlr) ) );

  subsyst = first_subsystem;

  while (subsyst != NULL &&
         SUBWORD_RULE != subsyst->get_r_type() )
    subsyst = subsyst->get_next_subsystem();

  if (subsyst == NULL) {
    cout << "No subword system in add_instances." << endl;
    if (get_echo() ) exit(1);
    return;
  }

  initialize_save_list(SUBWORD_RULE,1000,savell);

  nfail = 0;
  nrules = subsyst->get_nrules();
  orig = 0;
  initialize_loop();

  for (i = 0; i < nrules; i++) {
    count_loop();
    if (abort() ) break;
    subsyst->rule_by_origin(orig,p_rule);
    ll = p_rule.left.length;
    lr = p_rule.right.length;
    if (ll > maxill || lr > maxilr) continue;
    subsyst->unpack(p_rule,rul);
    promote(rul);

    rule_instances(rul,maxtll,maxtlr, maxlen,firstg,lastg,
                   nfail);
  }

  end_loop();

  cout << "The number of new rules is " << nsave << endl;
  cout << "The number of failures is " << nfail << endl;

  if (nsave > 0) {
    if (use_heuristics) heuristics();
    (void) reduce_all_subsystems(SUBWORD_RULE);
    append_save_list(subsyst);
    subsyst->make_index();
  }

  else append_save_list(subsyst); // This just clears save list

  return ;
}



/////////////////////////////////////////////////////////////////
//
//                rewriting_system::get_nonobvious
//
/////////////////////////////////////////////////////////////////

void rewriting_system::get_nonobvious(void)
{
  long              code,numequal,numbig,numsaved,numnotred;
  long              numout,nbits,kbmin,kbincr,kbmax,kbniter;
  long              niter;
  long              maxlen,flag;
  char              str[500];
  bool              changed,reduced,echo;
  external_form     ext;
  rule_type         r_type,save_type;
  rule_reference    exclude = {NULL,0};
  rule              rul,rul1;
  subsystem        *subsyst,*subsyst0;
  fstream           fstrm_in,fstrm_out;

  nsave_total = 0;
  echo = get_echo();

  numequal = numnotred = numbig = numsaved = numout = 0;
  rul.set_max_length(bsize);
  rul1.set_max_length(bsize);

 get_in_file:
  cout <<
    "\nEnter name of input file (or \"stop\"):  ";

  cin >> str;
  if (echo) cout << str << endl;
  if ( !cin.ignore(10000,'\n') ) goto input_error;

  if( 0 == strcmp(str,"stop") ) return;

  if (!file_exists(str) ) {
    cout << "File does not exist" << endl;
    if (echo) exit(1);
    goto get_in_file;
  }

  fstrm_in.open(str, ios::in);

 get_type:
  cout <<
    "Enter rule type (subword or prefix):  ";

  cin >> str;
  if (echo) cout << str << endl;
  if ( !cin.ignore(10000,'\n') ) goto input_error;

  if      (0 == strcmp(str,"subword") ) r_type = SUBWORD_RULE;
  else if (0 == strcmp(str,"prefix") )  r_type = PREFIX_RULE;

  else {
    cout << "Invalid type" << endl;
    if (echo) exit(1);
    goto get_type;
  }

  rul.r_type = r_type;
  rul1.r_type = r_type;

 get_format:
  nbits = n_bits;

  cout << "Enter external format" << endl <<
          "(names, len_names, len_numbers, or len_6bit):  ";

  cin >> str;
  if (echo) cout << str << endl;
  if ( !cin.ignore(10000,'\n') ) goto input_error;

  if (0 == strcmp(str,"names") )
    ext = NAMES;

  else if (0 == strcmp(str,"len_names") )
    ext = LEN_NAMES;

  else if (0 == strcmp(str,"len_numbers") )
    ext = LEN_NUMBERS;

  else if (0 == strcmp(str,"len_6bit") ) {
    ext = LEN_6BIT;
    cout << "\nEnter value of nbits:  ";
    scan_long(cin,nbits);
  }

  else {
    cout << "Invalid format" << endl;
    if (echo) exit(1);
    goto get_format;
  }

 get_out_file:
  cout <<
    "\nEnter name of output file (or \"stop\"):  ";

  cin >> str;
  if (echo) cout << str << endl;
  if ( !cin.ignore(10000,'\n') ) goto input_error;

  if (0 == strcmp(str,"stop") ) return;

  if (file_exists(str) ) {
    cout << "File exists already" << endl;
    if (echo) exit(1);
    goto get_out_file;
  }

  else fstrm_out.open(str, ios::out);

  // Get the kb parameters.

  cout << "\nEnter minimum kb overlap (positive integer):  ";
  cin  >> kbmin;
  if (echo) cout << kbmin << endl;
  if (!cin.ignore(10000,'\n') ) goto input_error;

  cout << "\nEnter kb overlap increment (positive integer):  ";
  cin  >> kbincr;
  if (echo) cout << kbincr << endl;
  if (!cin.ignore(10000,'\n') ) goto input_error;

  cout << "\nEnter maximum kb overlap (positive integer):  ";
  cin  >> kbmax;
  if (echo) cout << kbmax << endl;
  if (!cin.ignore(10000,'\n') ) goto input_error;

  cout << "\nEnter max saved left side (positive integer):  ";
  cin  >> savell;
  if (echo) cout << savell << endl;
  if ( !cin.ignore(10000,'\n') ) goto input_error;

  cout << "\nEnter max saved right side (nonnegative integer):  ";
  cin  >> savelr;
  if (echo) cout << savelr << endl;
  if (!cin.ignore(10000,'\n') ) goto input_error;

  set_save(savell,savelr);

  cout << "\nEnter number of kb iterations (integer):  ";
  cin >> kbniter;
  if (echo) cout << kbniter << endl;
  if ( !cin.ignore(10000,'\n') ) goto input_error;

  save_type = r_type;

  subsyst = first_subsystem;

  while (subsyst != NULL && save_type != subsyst->get_r_type() )
    subsyst = subsyst->get_next_subsystem();

  if (NULL == subsyst) {
    subsyst = new subsystem(this,save_type,HYBRID);
    add_subsystem(subsyst);
  }

  subsyst0 = subsyst;

  // 10 in next line should probably be 1.

  initialize_save_list(save_type,10,savell);

  if (!echo) initialize_loop();

  for (;;) {
    if (!echo) count_loop();
    if (abort() ) break;
    code = read_word(fstrm_in,ext,nbits,rul.left);
    if (1 == code) break;
    if (2 == code) goto input_error;

    code = read_word(fstrm_in,ext,nbits,rul.right);

    if (1 == code || 2 == code) goto input_error;

    rul1.left.set_copy(rul.left);
    rul1.right.set_copy(rul.right);
    rewrite(rul,exclude,changed,reduced);

    if (orient_rule(rul) ) {
      numequal++;
      continue;
    }

    if (!reduced) {
      numnotred++;
      continue;
    }

    balance_rule(rul);

    if (check_rule(rul) ) {
      save_rule(rul);
      numout++;
      print_word(fstrm_out,ext,rul1.left);
      fstrm_out << " ";
      print_word(fstrm_out,ext,rul1.right);
      fstrm_out << endl;
      numsaved++;
    }

    else {
      numbig++;
      continue;
    }

    if (use_heuristics) heuristics();
    reduce_all_subsystems(save_type);
    append_save_list(subsyst0);
    if (subsyst0->origin_in_use() ) subsyst0->make_origin();
    subsyst0->make_index();

    maxlen = rul.left.get_length() + kbincr;
    if (maxlen < kbmin) maxlen = kbmin;
    if (maxlen > kbmax) maxlen = kbmax;

    if (r_type == SUBWORD_RULE) flag = 5;
    else flag = 6;

    niter = kbniter;
    (void) kb_two_step(LONG_MAX,maxlen,niter,0,flag);
    initialize_save_list(save_type,10,savell);
  }

  if (!echo) end_loop();

  fstrm_in.close();
  fstrm_out.close();

  append_save_list(subsyst0);  // This just clears save list.  

  cout << "Number of equalities:  " << numequal << endl;
  cout << "Number not reduced:  " << numnotred << endl;
  cout << "Number too big:  " << numbig << endl;
  cout << "Number saved:  " << numsaved << endl;


  return;

 input_error:
  cout << "Input error" << endl;
  return;
}



/////////////////////////////////////////////////////////////////
//
//                 rewriting_system::geodesics
//
/////////////////////////////////////////////////////////////////

void rewriting_system::geodesics(
  vector_long    &wrd,
  long            maxlen
)
{
  long            i,j,k,ng,g,lw,status,origlen;
  double         *cnt,num,sum;
  bool            print,subword_only,echo,chg,red;
  char            str[100];
  vector_long     wrd1;
  rule_type       rtp;
  packed_rule     p_rule;
  rule_reference  exclude = {NULL,0};
  external_form   ext;
  subsystem      *subsyst;
  fstream         fstrm_out;

  update_time();
  echo = get_echo();

  subword_only = TRUE;

  cout << "\nDo you want the words written to a file:  ";

  if (YES) {
    print = TRUE;

   get_file:
    cout << "\nEnter name of file (or \"stop\"):  ";
    cin >> str;
    if (echo) cout << str << endl;

    if ( !cin.ignore(10000,'\n') ) goto input_error;
    if ( 0 == strcmp(str,"stop") ) return;

    if (file_exists(str) ) {
      cout << "File does not exist" << endl;
      if (echo) exit(1);
      goto get_file;
    }

    fstrm_out.open(str, ios::out);

   get_format:
    cout << "Enter external format" << endl <<
            "(names, len_names, len_numbers, or len_6bit):  ";

    cin >> str;
    if (echo) cout << str << endl;
    if ( !cin.ignore(10000,'\n') ) goto input_error;

    if (0 == strcmp(str,"names") )
      ext = NAMES;

    else if (0 == strcmp(str,"len_names") )
      ext = LEN_NAMES;

    else if (0 == strcmp(str,"len_numbers") )
      ext = LEN_NUMBERS;

    else if (0 == strcmp(str,"len_6bit") )
      ext = LEN_6BIT;

    else {
      cout << "Invalid format" << endl;
      if (echo) exit(1);
      goto get_format;
    }
  }

  else print = FALSE;

  origlen = wrd.get_length();
  wrd.set_max_length(maxlen);
  wrd1.set_max_length(maxlen);

  cnt = (double *) fxmllc( (maxlen+1)*sizeof(double) );

  if (cnt == NULL) {
    cout << "Unable to allocate space for counts" << endl;
    return;
  }

  for (j = 0; j <= maxlen; j++) cnt[j] = 0.0;

  rewrite_from_left(TRUE,wrd,wrd1,exclude,chg,red);

  if (wrd1.get_length() < origlen) {
    cout << "The prefix is not a geodesic." << endl;
    return;
  }

  cnt[origlen] = 1.0;

  if (maxlen <= origlen) goto done;

  // Begin backtrack search for nonempty geodesics.

  lw = origlen + 1;
  wrd.set_length(lw);
  wrd.set_entry(lw-1,0);

  for (;;) {
    rewrite_from_left(TRUE,wrd,wrd1,exclude,chg,red);

    if (wrd1.get_length() < wrd.get_length() ) goto increment;

    cnt[lw] += 1.0;

    if (print) {
      print_word(fstrm_out,ext,wrd);
      fstrm_out << endl;
    }

    if (lw == maxlen) goto increment;

    wrd.set_entry(lw++,0);
    wrd.set_length(lw);
    continue;

   increment:
    while (lw > origlen && (g = wrd[lw-1]) == t_gens-1) lw--;
    if (lw <= origlen) break;
    wrd.set_entry(lw-1,g+1);
    wrd.set_length(lw);
  }
    
 done:
  sum = 0.0;
  cout << "The numbers of geodesics are:" << endl << endl;

  for (i = origlen; i <= maxlen; i++) {
    cout << setw(4) << i << "  " << setw(12) << setprecision(0)
         << setiosflags(ios::fixed) << cnt[i] << endl;
    sum += cnt[i];

    if (cnt[i] < 1.0) break;
  }
  
  cout << endl << endl;
  cout << "The total number of geodesics is " << sum << endl;

  fxfree( (void *)cnt );

  return;

 input_error:
  cout << "Error clearing line" << endl;
  exit(1);
}



/////////////////////////////////////////////////////////////////
//
//                 rewriting_system::count_subgroup
//
/////////////////////////////////////////////////////////////////

void rewriting_system::count_subgroup(void)
{
  long            i,j,k,ng,g,lw,status,origlen,ngens,nelts;
  long            maxlen,s,t,u,code,first,last,nbits;
  long            maxnelts,hprime,maxorig,m,lw2,maxlw;
  double         *cnt,num,sum;
  bool            print,subword_only,echo,changed,reduced;
  bool            print_table;
  char            str[100];
  rule_type       rtp;
  packed_rule     p_rule;
  rule_reference  exclude = {NULL,0};
  external_form   ext;
  vector_long     gens[20];
  uchar          *elts;
  long            origin[500001];
  ulong           elthash[700001];
  vector_long     wrd;
  subsystem      *subsyst;
  fstream         fstrm_out,fstrm_out2;

  maxnelts = 500000;
  hprime   = 700001;
  maxorig  = 5000000;

  elts = (uchar *) fxmllc( maxorig);

  update_time();
  echo = get_echo();
  wrd.set_max_length(2000);

  cout << "Enter number of subgroup generators:  ";
  cin  >> ngens;
  if (echo) cout << ngens << endl;

  if (ngens > 20) {
    cout << "Too many generators" << endl;
    goto finished;
  }

  cout << "Enter the generators is names format:" << endl;

  nbits = 0;

  for (i = 0; i < ngens; i++) {
    code = read_word(cin,NAMES,nbits,gens[i]);

    if (code == 1 || code == 2) {
      cout << "Trouble reading generator" << endl;
      goto finished;
    }
  }

  cout << "Enter maximum length:  ";
  cin >> maxlen;
  if (echo) cout << maxlen << endl;

  subword_only = TRUE;

  cout << "\nDo you want the words written to a file:  ";

  if (YES) {
    print = TRUE;

   get_file:
    cout << "\nEnter name of file (or \"stop\"):  ";
    cin >> str;
    if (echo) cout << str << endl;

    if ( !cin.ignore(10000,'\n') ) goto input_error;
    if ( 0 == strcmp(str,"stop") ) goto finished;

    if (file_exists(str) ) {
      cout << "File exists" << endl;
      if (echo) exit(1);
      goto get_file;
    }

    fstrm_out.open(str, ios::out);

   get_format:
    cout << "Enter external format" << endl <<
            "(names, len_names, len_numbers, or len_6bit):  ";

    cin >> str;
    if (echo) cout << str << endl;
    if ( !cin.ignore(10000,'\n') ) goto input_error;

    if (0 == strcmp(str,"names") )
      ext = NAMES;

    else if (0 == strcmp(str,"len_names") )
      ext = LEN_NAMES;

    else if (0 == strcmp(str,"len_numbers") )
      ext = LEN_NUMBERS;

    else if (0 == strcmp(str,"len_6bit") )
      ext = LEN_6BIT;

    else {
      cout << "Invalid format" << endl;
      if (echo) exit(1);
      goto get_format;
    }
  }

  else print = FALSE;

  cout << "\nDo you want the coset table written to a file:  ";

  if (YES) {
    print_table = TRUE;

   get_file2:
    cout << "\nEnter name of file (or \"stop\"):  ";
    cin >> str;
    if (echo) cout << str << endl;

    if ( !cin.ignore(10000,'\n') ) goto input_error;
    if ( 0 == strcmp(str,"stop") ) goto finished;

    if (file_exists(str) ) {
      cout << "File exists" << endl;
      if (echo) exit(1);
      goto get_file;
    }

    fstrm_out2.open(str, ios::out);
  }

  else print_table = FALSE;

  origin[0] = 0;
  origin[1] = 0;
  for (i = 0; i < hprime; i++) elthash[i] = maxnelts + 10;  
  nelts = 1;
  elthash[0] = 0;
  maxlw = 0;

  cnt = (double *) fxmllc( (maxlen+1)*sizeof(double) );

  if (cnt == NULL) {
    cout << "Unable to allocate space for counts" << endl;
    goto finished;
  }

  cnt[0] = 1.0;
  for (j = 1; j <= maxlen; j++) cnt[j] = 0.0;

  first = 0;
  last = 0;

  for (i = 1; i <= maxlen; i++) {
    for (j = first; j <= last; j++) {
      if (print_table) fstrm_out2 << j+1 << " ";

      for (k = 0; k < ngens; k++) {
        lw = origin[j+1] - origin[j];
        wrd.set_length(lw);
        t = origin[j];
        for (s = 0; s < lw; s++) wrd.safe_set_entry(s,elts[t++]);
        wrd.append(gens[k]);
        rewrite_from_left(subword_only,wrd,wrd,exclude,changed,reduced);

        lw = wrd.get_length();
        m = 0;

        for (s = 0; s < lw; s++)
          m = (11*m + wrd.safe_get_entry(s) + 1) % hprime;

        while (elthash[m] < maxnelts) {
          s = elthash[m];
          lw2 = origin[s+1] - origin[s];
          u = origin[s];

          if (lw2 != lw) goto unequal;

          for (t = 0; t < lw; t++)
            if (elts[u++] != wrd.safe_get_entry(t) ) goto unequal;

          if (print_table) fstrm_out2 << " " << s+1;

          break;

         unequal:
          m++;
          if (m == hprime) m = 0;
        }

        if (elthash[m] > maxnelts) {
          if (nelts >= maxnelts) continue;
          if (origin[nelts]+lw >= maxorig) continue;

          if (lw > maxlw) maxlw = lw;
          s = origin[nelts];
          elthash[m] = nelts++;
          origin[nelts] = s + lw;
          if (print_table) fstrm_out2 << " " << nelts;

          for (t = 0; t < lw; t++) elts[s++] = wrd.safe_get_entry(t);

          if (print) {
            print_word(fstrm_out,ext,wrd);
            fstrm_out << endl;
          }
        }
      }

      if (print_table) fstrm_out2 << endl;
    }

    cnt[i] = nelts - last - 1;
    first = last + 1;
    last = nelts - 1;
  }

    
 done:
  sum = 0.0;
  cout << "The numbers of irreducible words are:" << endl << endl;

  for (i = 0; i <= maxlen; i++) {
    cout << setw(4) << i << "  " << setw(12) << setprecision(0)
         << setiosflags(ios::fixed) << cnt[i] << endl;
    sum += cnt[i];

    if (cnt[i] < 1.0) break;
  }
  
  cout << endl << endl;
  cout << "The total number of words is " << sum << endl;

  fxfree( (void *)cnt );

  cout << endl << " The maximum length in the overgroup is "
       << maxlw << endl;


 finished:
  fxfree( (void*) elts);
  if (print) fstrm_out.close();
  if (print_table) fstrm_out.close();

  return;

 input_error:
  cout << "Error clearing line" << endl;
  exit(1);
}



/////////////////////////////////////////////////////////////////
//
//                 rewriting_system::shift_rules
//
/////////////////////////////////////////////////////////////////

void rewriting_system::shift_rules(
  long  llmax,  // Maximum initial left side.
  long  lrmax,  // Maximum initial right side.
  long  firstg,
  long  shift
)
{
  long          nrules,orig,i,j,ll,lr;
  ulong         numequal,numnotred,numsaved,numbig;
  ulong         x;
  bool          changed,reduced,echo;
  packed_rule   p_rule;
  rule          rul;
  rule_reference exclude = {NULL,0};
  subsystem    *subsyst,*subsyst1;

  echo = get_echo();

  if (0 >= firstg || firstg >= t_gens || shift < 1 || shift > firstg) {
    cout << "Error in arguments" << endl;
    if (get_echo() ) exit(1);
    return;
  }

  if (print_level >= 1) {
    cout << "Shifting subword rules"
         << ", firstg = " << firstg << ", shift = " << shift << endl;

    if (logll > 0)
      log_stream << "Shifting subword rules"
         << ", firstg = " << firstg << ", shift = " << shift << endl;
  }

  nsave_total = 0;
  update_time();

  rul.set_max_length(bsize);

  subsyst = first_subsystem;

  while (subsyst != NULL &&
         SUBWORD_RULE != subsyst->get_r_type() )
    subsyst = subsyst->get_next_subsystem();

  if (subsyst == NULL) {
    cout << "No subword system in shift_rules." << endl;
    if (get_echo() ) exit(1);
    return;
  }

  initialize_save_list(SUBWORD_RULE,1000,savell);

  numequal  = 0;
  numnotred = 0;
  numsaved  = 0;
  numbig    = 0;
  nrules = subsyst->get_nrules();
  orig = 0;
  if (!echo) initialize_loop();

  for (i = 0; i < nrules; i++) {
    if (!echo) count_loop();
    if (abort() ) break;
    subsyst->rule_by_origin(orig,p_rule);
    ll = p_rule.left.length;
    lr = p_rule.right.length;
    if (ll > llmax || lr > lrmax) continue;
    subsyst->unpack(p_rule,rul);

    for (j = 0; j < ll; j++) {
      x = rul.left[j];
      if (x < firstg) break;
      rul.left.set_entry(j,x-shift);
    }

    if (j < ll) continue;

    for (j = 0; j < lr; j++) {
      x = rul.right[j];
      if (x < firstg) break;
      rul.right.set_entry(j,x-shift);
    }

    if (j < lr) continue;

    rewrite(rul,exclude,changed,reduced);

    if (orient_rule(rul) ) {
      numequal++;
      continue;
    }

    if (!reduced) {
      numnotred++;
      continue;
    }

    balance_rule(rul);

    if (check_rule(rul) ) {
      save_rule(rul);
      numsaved++;
    }

    else numbig++;
  }

  if (!echo) end_loop();

  cout << "The number of new rules is " << numsaved << endl;
  cout << "The number not reduced is " << numnotred << endl;
  cout << "The number equal is " << numequal << endl;
  cout << "The number too big is " << numbig << endl;

  if (nsave > 0) {
    if (use_heuristics) heuristics();
    (void) reduce_all_subsystems(SUBWORD_RULE);
    append_save_list(subsyst);
    subsyst->make_index();
  }

  else append_save_list(subsyst); // This just clears save list

  return ;
}





// End of system.C
